# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

from mdo_engine.boundary import DataDefinition
from mdo_engine.entity.data import MetaData


class CoreMetaData(MetaData):
    """Concrete MetaData class for storing metadata for a variable in memory."""

    def __init__(self, props_dict):
        self._title = None
        self._description = None
        self._category = None
        self._group = None
        self._auto_only = None
        self._types = None
        self._tables = None
        self._units: Optional[list[str]] = None
        self._labels = None
        self._valid_values = None
        self._minimums = None
        self._minimum_equals = None
        self._maximums = None
        self._maximum_equals = None
        self._experimental = None

        super(CoreMetaData, self).__init__(props_dict)

    @classmethod
    def get_base_properties(cls):
        """Properties that must be provided"""
        base_props = super(CoreMetaData, cls).get_base_properties()
        base_props.extend(["title"])
        return base_props

    @property
    def title(self):
        """A short name for the data"""

        if self.experimental is None:
            return self._title

        if True in self.experimental:
            title = "{} (Experimental)".format(self._title)
        else:
            title = self._title

        return title

    @property
    def description(self):
        """A longer description of the data"""
        return self._description

    @property
    def category(self):
        """Highest categorisation. Can be none if contained in variable id"""
        return self._category

    @property
    def group(self):
        """Optional grouping of variables below category or subcategory"""
        return self._group

    @property
    def auto_only(self):
        """Call auto_db method only if this variable has a value"""
        return self._auto_only

    @property
    def types(self):
        """A list of data type can also be specified which may be used
        within the data structure"""
        return self._types

    @property
    def tables(self):
        """List of database tables used as an interface to the data."""
        return self._tables

    @property
    def units(self):
        """The SI units of the data"""
        return self._units

    @property
    def labels(self):
        """Short names for dimensions of data"""
        return self._labels

    @property
    def minimums(self):
        """Data must be greater than this value"""
        return self._minimums

    @property
    def minimum_equals(self):
        """Data must be greater than or equal to this value"""
        return self._minimum_equals

    @property
    def maximums(self):
        """Data must be less than this value"""
        return self._maximums

    @property
    def maximum_equals(self):
        """Data must be less than or equal to this value"""
        return self._maximum_equals

    @property
    def valid_values(self):
        """List of valid values for discrete data"""
        return self._valid_values

    @property
    def experimental(self):
        """List of experimental values or entire variable if list contains
        True"""
        return self._experimental


class CoreData(DataDefinition):
    def __init__(self):
        super(CoreData, self).__init__()

    @property
    def package_name(self):
        return "dtocean-core"

    @property
    def company_name(self):
        return "DTOcean"

    @property
    def local_yaml_dir(self):
        """The paths of the yaml definition files."""
        return "yaml"

    @property
    def user_yaml_dir(self):
        """The paths of the yaml definition files."""
        return None



import os
from typing import Any

from dtocean_core.utils.database import (
    MIN_DB_VERSION,
    database_from_files,
    database_to_files,
    draw_map,
    filter_map,
    get_database,
    get_database_config,
    get_table_map,
)

from .shared import SmartFormatter


def subcommand(subparser):
    desStr = "Convert DTOcean database to and from structured files"
    parser = subparser.add_parser(
        "database",
        description=desStr,
        help="export and import database",
        formatter_class=SmartFormatter,
    )
    _setup_database(parser)


def _setup_database(parser):
    parser.add_argument(
        "action",
        choices=["dump", "load", "list", "view", "dir"],
        help="R|Select an action, where\n"
        " dump = export database to files\n"
        " load = import files into database\n"
        " list = print stored credentials identifiers\n"
        " view = print stored credentials (using -i "
        "option)\n"
        "  dir = print table structure",
    )

    parser.add_argument(
        "-d",
        "--directory",
        help=("directory to add or read files from. " "Defaults to '.'"),
        type=str,
        default=".",
    )

    parser.add_argument(
        "-s",
        "--section",
        choices=["device", "site", "other"],
        help="R|Only operate on section from\n"
        " device = tables related to the OEC\n"
        " site = tables related to the deployment site\n"
        " other = tables related to the reference data",
    )

    parser.add_argument(
        "-i", "--identifier", help=("stored credentials identifier"), type=str
    )

    parser.add_argument("--host", help=("override database host"), type=str)

    parser.add_argument("--name", help=("override database name"), type=str)

    parser.add_argument("--user", help=("override database username"), type=str)

    parser.add_argument(
        "-p", "--pwd", help=("override database password"), type=str
    )

    parser.set_defaults(func=lambda args: _database(args))


def _database(args):
    """Command line interface for database_to_files and database_from_files.

    Example:

        To get help::

            dtocean-database -h

    """

    cred: dict[str, Any]
    _, config = get_database_config()

    request = {
        "action": args.action,
        "root_path": args.directory,
        "filter_table": args.section,
        "db_id": args.identifier,
        "db_host": args.host,
        "db_name": args.name,
        "db_user": args.user,
        "db_pwd": args.pwd,
    }

    # List the available database configurations
    if request["action"] == "list":
        id_str = ", ".join(config.keys())

        if id_str:
            msg_str = (
                "Available database configuration identifiers are: " "{}"
            ).format(id_str)
        else:
            msg_str = "No database configurations are stored"

        print(msg_str)

        return

    if request["action"] == "view":
        if request["db_id"] is None:
            err_msg = "Option '-i' must be specified with 'view' action"
            raise ValueError(err_msg)

        cred = config[request["db_id"]]

        for k, v in cred.items():
            print("{:>8} ::  {}".format(k, v))

        return

    table_list = get_table_map()

    # Filter the table if required
    if request["filter_table"] is not None:
        filtered_dict = filter_map(table_list, request["filter_table"])
        table_list = [filtered_dict]

    if request["action"] == "dir":
        print("\n" + draw_map(table_list))
        return

    # Set up the DB
    if request["db_id"] is not None:
        cred = config[request["db_id"]]
    else:
        cred = {"host": None, "dbname": None, "user": None, "pwd": None}

    if request["db_host"] is not None:
        cred["host"] = request["db_host"]

    if request["db_name"] is not None:
        cred["dbname"] = request["db_name"]

    if request["db_user"] is not None:
        cred["user"] = request["db_user"]

    if request["db_pwd"] is not None:
        cred["pwd"] = "postgres"

    db = get_database(cred, timeout=60, min_version=MIN_DB_VERSION)

    if request["action"] == "dump":
        # make a directory if required
        if not os.path.exists(request["root_path"]):
            os.makedirs(request["root_path"])

        database_to_files(
            request["root_path"],
            table_list,
            db,
            print_function=print,
        )

        return

    if request["action"] == "load":
        database_from_files(
            request["root_path"],
            table_list,
            db,
            print_function=print,
        )

        return

    raise RuntimeError("Highly illogical...")



# -*- coding: utf-8 -*-

#    Copyright (C) 2020 National Technology & Engineering Solutions of Sandia
#    Copyright (C) 2017-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Created on Mon Sep 11 08:49:07 2017

.. moduleauthor:: Mathew Topper <damm_horse@yahoo.co.uk>
"""

import logging
from typing import TypeAlias, cast

import numpy as np
import numpy.typing as npt
from contourpy import LineType, contour_generator
from scipy import optimize, stats
from scipy.special import gamma

PointArray: TypeAlias = npt.NDArray[np.float64]

# Set up logging
module_logger = logging.getLogger(__name__)


class UniVariateKDE:
    def __init__(self, data, bandwidth=0.3):
        self._kde = stats.gaussian_kde(data, bw_method=bandwidth)
        self._cdf = None
        self._ppf = None
        self._x0 = 0.0

    def pdf(self, values):
        return self._kde(values)

    def cdf(self, values):
        if self._cdf is None:
            self._cdf = self._calc_cdf()

        return self._cdf(values)

    def ppf(self, probabilities, x0=None):
        if self._ppf is None or self._x0 != x0:
            self._ppf = self._calc_ppf(x0)
            self._x0 = x0

        result = self._ppf(probabilities)

        if np.isnan(result).any():
            result = None

        return result

    def mean(self):
        return self._kde.dataset.mean()

    def median(self):
        return np.median(self._kde.dataset)

    def mode(self, samples=1000):
        """Numerically search for the mode"""

        x = np.linspace(
            self._kde.dataset.min(), self._kde.dataset.max(), samples
        )
        most_likely = x[np.argsort(self._kde(x))[-1]]

        return most_likely

    def confidence_interval(self, percent, x0=None):
        if self._ppf is None:
            self._ppf = self._calc_ppf(x0)

        x = percent / 100.0
        bottom = (1 - x) / 2
        top = (1 + x) / 2

        result = self._ppf([bottom, top])

        if np.isnan(result).any():
            result = None

        return result

    def _calc_cdf(self):
        def _kde_cdf(x):
            return self._kde.integrate_box_1d(-np.inf, x)

        kde_cdf = np.vectorize(_kde_cdf)

        return kde_cdf

    def _calc_ppf(self, x0=None):
        if self._cdf is None:
            self._cdf = self._calc_cdf()

        def _kde_ppf(q):
            def f(x, q):
                assert self._cdf is not None
                return self._cdf(x) - q

            x0_list = [
                self.mean(),
                0.0,
                self._kde.dataset.min(),
                self._kde.dataset.max(),
            ]

            if x0 is not None:
                x0_list = [x0] + x0_list

            for x0_local in x0_list:
                result = optimize.fsolve(
                    f,
                    x0_local,
                    args=(q,),
                    full_output=True,
                )

                if result[2] == 1:
                    return result[0][0]

            return np.nan

        kde_ppf = np.vectorize(_kde_ppf)

        return kde_ppf


class BiVariateKDE:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.kernel = self._set_kernel()

    def _set_kernel(self):
        values = np.vstack([self.x, self.y])
        return stats.gaussian_kde(values)

    def mean(self):
        return self.x.mean(), self.y.mean()

    def median(self):
        return np.median(self.kernel.dataset, 1)

    def mode(self, xtol=0.0001, ftol=0.0001, disp=False):
        """Determine the ordinate of the most likely value of the given KDE"""

        median = self.median()
        modal_coords = optimize.fmin(
            lambda x: -1 * self.kernel(x),
            median,
            xtol=xtol,
            ftol=ftol,
            disp=disp,
        )

        return modal_coords

    def pdf(self, x_range=None, y_range=None, npoints=1000):
        # Wide estimate on the ranges if not given
        if x_range is None:
            dx = self.x.max() - self.x.min()
            x_range = (self.x.min() - dx, self.x.max() + dx)

        if y_range is None:
            dy = self.y.max() - self.y.min()
            y_range = (self.y.min() - dy, self.y.max() + dy)

        X, Y = np.mgrid[
            x_range[0] : x_range[1] : (npoints * 1j),
            y_range[0] : y_range[1] : (npoints * 1j),
        ]
        positions = np.vstack([X.ravel(), Y.ravel()])

        xx = X[:, 0]
        yy = Y[0, :]
        pdf = np.reshape(self.kernel(positions).T, X.shape)

        return xx, yy, pdf


def pdf_confidence_densities(pdf, levels=None):
    """Determine the required density values to satisfy a list of confidence
    levels in the given pdf"""

    def diff_frac(density, pdf, target_frac, pdf_sum):
        density_frac = pdf[pdf >= density].sum() / pdf_sum
        return density_frac - target_frac

    if levels is None:
        levels = np.array([95.0])
    else:
        levels = np.array(levels)

    fracs = levels / 100.0
    pdf_sum = pdf.sum()
    densities = []

    for frac in fracs:
        local_pdf = np.copy(pdf)

        try:
            density = optimize.brentq(
                diff_frac, pdf.min(), pdf.max(), args=(local_pdf, frac, pdf_sum)
            )

            densities.append(density)
        except ValueError as e:
            module_logger.debug(e, exc_info=True)

    return densities


def pdf_contour_coords(xx, yy, pdf, level):
    cont_gen = contour_generator(xx, yy, pdf.T, line_type=LineType.Separate)
    lines = cast(list[PointArray], cont_gen.lines(level))

    cx = []
    cy = []

    for v in lines:
        cx.extend(v[:, 0])
        cy.extend(v[:, 1])

    return cx, cy


def get_standard_error(values):
    """
    Calculates the standard error of the mean of a given function.

    This function can either be used to reduce the standard error to
    a certain level or calculate the standard error given a fixed number
    of samples.

    Args:
      values

    Returns: the standard error metric vector
    """

    def get_c4(n):
        # Correction for unbiased estimate of the standard deviation.
        # https://en.wikipedia.org/wiki/Unbiased_estimation_of_standard_deviation
        b_bottom = gamma((n - 1) / 2.0)

        if np.isinf(b_bottom):
            return np.inf

        a = np.sqrt(2.0 / (n - 1))
        b_top = gamma(n / 2.0)
        return a * b_top / b_bottom

    n = len(values)

    if n < 2:
        return None

    c4 = get_c4(n)

    if not np.isfinite(c4):
        c4 = 1

    result_error = c4 * np.std(values) / np.sqrt(n)

    return result_error



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from mdo_engine.boundary.interface import MaskVariable, QueryInterface

from . import ProjectInterface


class SystemTypeInterface(ProjectInterface):
    """Interface for ensuring the system type has been provided"""

    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "System Type Selection"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = ["device.system_type"]

        return input_list

    @classmethod
    def declare_outputs(cls):
        """A class method to declare all the output variables provided by
        this interface.

        Returns:
          list: List of output identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              outputs = ["My:first:variable",
                         "My:third:variable",
                        ]
        """

        outputs = ["hidden.pipeline_active"]

        return outputs

    @classmethod
    def declare_optional(cls):
        """A class method to declare all the variables which should be flagged
        as optional.

        Returns:
          list: List of optional variable identifiers

        Note:
          Currently only inputs marked as optional have any logical effect.
          However, this may change in future releases hence the general
          approach.

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the declare_inputs output. For example::

              optional = ["My:first:variable",
                         ]
        """

        return None

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example:

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "not_used": "device.system_type",
            "output": "hidden.pipeline_active",
        }

        return id_map

    def connect(self):
        """The connect method is used to execute the external program and
        populate the interface data store with values.

        Note:
          Collecting data from the interface for use in the external program
          can be accessed using self.data.my_input_variable. To put new values
          into the interface once the program has run we set
          self.data.my_output_variable = value

        """

        self.data.output = True


class OptionsInterface(ProjectInterface):
    """Interface for providing database filtering option values"""

    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Site and System Options"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = [
            "device.system_type",
            "hidden.available_sites",
            "hidden.available_systems",
            "hidden.corridor_boundaries",
            "hidden.lease_boundaries",
            "hidden.site_boundaries",
            "hidden.landing_points",
        ]

        return input_list

    @classmethod
    def declare_outputs(cls):
        """A class method to declare all the output variables provided by
        this interface.

        Returns:
          list: List of output identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              outputs = ["My:first:variable",
                         "My:third:variable",
                        ]
        """

        output_list = [
            "device.available_names",
            "site.available_names",
            "hidden.corridor_selected",
            "hidden.lease_selected",
        ]

        return output_list

    @classmethod
    def declare_optional(cls):
        """A class method to declare all the variables which should be flagged
        as optional.

        Returns:
          list: List of optional variable identifiers

        Note:
          Currently only inputs marked as optional have any logical effect.
          However, this may change in future releases hence the general
          approach.

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the declare_inputs output. For example::

              optional = ["My:first:variable",
                         ]
        """

        option_list = [
            "hidden.available_sites",
            "hidden.available_systems",
            "hidden.corridor_boundaries",
            "hidden.lease_boundaries",
            "hidden.site_boundaries",
            "hidden.landing_points",
        ]

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example:

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "sys_type": "device.system_type",
            "all_sites": "hidden.available_sites",
            "all_systems": "hidden.available_systems",
            "corridor_boundaries": "hidden.corridor_boundaries",
            "corridor_selected": "hidden.corridor_selected",
            "lease_boundaries": "hidden.lease_boundaries",
            "site_names": "site.available_names",
            "site_boundaries": "hidden.site_boundaries",
            "landing_points": "hidden.landing_points",
            "system_names": "device.available_names",
            "lease_selected": "hidden.lease_selected",
        }

        return id_map

    def connect(self):
        """The connect method is used to execute the external program and
        populate the interface data store with values.

        Note:
          Collecting data from the interface for use in the external program
          can be accessed using self.data.my_input_variable. To put new values
          into the interface once the program has run we set
          self.data.my_output_variable = value

        """

        if self.data.all_systems is not None:
            systems_df = self.data.all_systems
            systems_type_df = systems_df[
                systems_df["device_type"].str.contains(self.data.sys_type)
            ]

            system_names = systems_type_df["description"]
            self.data.system_names = system_names.reset_index(drop=True)

        if self.data.all_sites is not None:
            self.data.site_names = self.data.all_sites["site_name"]
            self.data.corridor_selected = False
            self.data.lease_selected = False


class SiteBoundaryInterface(ProjectInterface):
    """Interface for populating cable corridor & lease area polygons and
    the projection string
    """

    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Site Boundary Selection"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = [
            "hidden.available_sites",
            "hidden.corridor_boundaries",
            "hidden.lease_boundaries",
            "hidden.landing_points",
            "site.selected_name",
            "hidden.site_boundaries",
        ]

        return input_list

    @classmethod
    def declare_outputs(cls):
        """A class method to declare all the output variables provided by
        this interface.

        Returns:
          list: List of output identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              outputs = ["My:first:variable",
                         "My:third:variable",
                        ]
        """

        output_list = [
            "site.corridor_boundary",
            "site.lease_boundary",
            "site.projection",
            "hidden.site_boundary",
            "corridor.landing_point",
            "hidden.corridor_selected",
            "hidden.lease_selected",
        ]

        return output_list

    @classmethod
    def declare_optional(cls):
        """A class method to declare all the variables which should be flagged
        as optional.

        Returns:
          list: List of optional variable identifiers

        Note:
          Currently only inputs marked as optional have any logical effect.
          However, this may change in future releases hence the general
          approach.

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the declare_inputs output. For example::

              optional = ["My:first:variable",
                         ]
        """

        option_list = None

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example:

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "all_sites": "hidden.available_sites",
            "corridor_poly": "site.corridor_boundary",
            "corridor_boundaries": "hidden.corridor_boundaries",
            "corridor_selected": "hidden.corridor_selected",
            "landing_points": "hidden.landing_points",
            "landing_point": "corridor.landing_point",
            "lease_boundaries": "hidden.lease_boundaries",
            "lease_poly": "site.lease_boundary",
            "proj_string": "site.projection",
            "selected_site": "site.selected_name",
            "lease_selected": "hidden.lease_selected",
            "site_boundaries": "hidden.site_boundaries",
            "site_poly": "hidden.site_boundary",
        }

        return id_map

    def connect(self):
        """The connect method is used to execute the external program and
        populate the interface data store with values.

        Note:
          Collecting data from the interface for use in the external program
          can be accessed using self.data.my_input_variable. To put new values
          into the interface once the program has run we set
          self.data.my_output_variable = value

        """

        sites_df = self.data.all_sites
        site_matches = sites_df["site_name"] == self.data.selected_site

        if np.sum(site_matches) == 0:
            names_str = ", ".join(sites_df["site_name"])
            errStr = (
                "No site with name '{}' found. Available names " "are {}"
            ).format(self.data.selected_site, names_str)
            raise ValueError(errStr)

        if np.sum(site_matches) > 1:
            errStr = (
                "More than one site with name '{}' found. Database "
                "is corrupt."
            ).format(self.data.selected_site)
            raise ValueError(errStr)

        site_corridor = self.data.corridor_boundaries[self.data.selected_site]
        site_lease = self.data.lease_boundaries[self.data.selected_site]
        site_boundary = self.data.site_boundaries[self.data.selected_site]
        landing_point = self.data.landing_points[self.data.selected_site]
        proj_strings = sites_df[site_matches]["lease_area_proj4_string"]

        self.data.corridor_poly = site_corridor
        self.data.lease_poly = site_lease
        self.data.site_poly = site_boundary
        self.data.landing_point = landing_point

        self.data.corridor_selected = True
        self.data.lease_selected = True

        self.data.proj_string = proj_strings.values[0]


class FilterInterface(ProjectInterface, QueryInterface):
    """Interface to filter the database for a selected site and technology."""

    def __init__(self):
        ProjectInterface.__init__(self)
        QueryInterface.__init__(self)

    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Database Filtering Interface"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = [
            "hidden.available_sites",
            "hidden.available_systems",
            "hidden.corridor_selected",
            "hidden.lease_selected",
            MaskVariable("site.selected_name", "hidden.available_sites"),
            MaskVariable("device.selected_name", "hidden.available_systems"),
        ]

        return input_list

    @classmethod
    def declare_outputs(cls):
        """A class method to declare all the output variables provided by
        this interface.

        Returns:
          list: List of output identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              outputs = ["My:first:variable",
                         "My:third:variable",
                        ]
        """

        outputs = ["hidden.site_filtered", "hidden.device_filtered"]

        return outputs

    @classmethod
    def declare_optional(cls):
        """A class method to declare all the variables which should be flagged
        as optional.

        Returns:
          list: List of optional variable identifiers

        Note:
          Currently only inputs marked as optional have any logical effect.
          However, this may change in future releases hence the general
          approach.

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the declare_inputs output. For example:

              optional = ["My:first:variable",
                         ]
        """

        option_list = [
            "hidden.available_sites",
            "hidden.available_systems",
            "hidden.corridor_selected",
            "hidden.lease_selected",
            "site.selected_name",
            "device.selected_name",
        ]

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example:

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "all_sites": "hidden.available_sites",
            "all_systems": "hidden.available_systems",
            "corridor_selected": "hidden.corridor_selected",
            "device_filtered": "hidden.device_filtered",
            "selected_site": "site.selected_name",
            "selected_system": "device.selected_name",
            "site_filtered": "hidden.site_filtered",
            "lease_selected": "hidden.lease_selected",
        }

        return id_map

    def connect(self):
        """The connect method is used to execute the external program and
        populate the interface data store with values.

        Note:
          Collecting data from the interface for use in the external program
          can be accessed using self.data.my_input_variable. To put new values
          into the interface once the program has run we set
          self.data.my_output_variable = value

        """

        # The main task is to prepare the database for providing project
        # data. This is done using stored proceedure calls. Firstly, get the
        # ID given the name

        if self._db is None:
            raise RuntimeError("No database connected")

        if self.data.selected_system is not None:
            systems_df = self.data.all_systems
            system_matches = (
                systems_df["description"] == self.data.selected_system
            )
            system_ids = systems_df[system_matches]["id"]

            if len(system_ids) == 0:
                names_str = ", ".join(systems_df["description"])
                errStr = (
                    "No system with name '{}' found. Available names " "are {}"
                ).format(self.data.selected_system, names_str)
                raise ValueError(errStr)

            if len(system_ids) > 1:
                errStr = (
                    "More than one system with name '{}' found. "
                    "Database is corrupt."
                ).format(self.data.selected_system)
                raise ValueError(errStr)

            self._db.call_function(
                "filter.sp_filter_device_data",
                system_ids.values[0].item(),
            )

            self.data.device_filtered = True

        if self.data.selected_site is not None:
            sites_df = self.data.all_sites
            site_matches = sites_df["site_name"] == self.data.selected_site
            site_ids = sites_df[site_matches]["id"]

            if len(site_ids) == 0:
                names_str = ", ".join(sites_df["site_name"])
                errStr = (
                    "No site with name '{}' found. Available names " "are {}"
                ).format(self.data.selected_system, names_str)
                raise ValueError(errStr)

            if len(site_ids) > 1:
                errStr = (
                    "More than one site with name '{}' found. Database "
                    "is corrupt."
                ).format(self.data.selected_site)
                raise ValueError(errStr)

            self._db.call_function(
                "filter.sp_filter_site_data",
                site_ids.values[0].item(),
            )

            self.data.site_filtered = True


class BoundariesInterface(ProjectInterface):
    """Interface to expose the area defining polygons."""

    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Project Boundaries Interface"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = [
            "site.projection",
            "site.lease_boundary",
            "site.corridor_boundary",
            "corridor.landing_point",
        ]

        return input_list

    @classmethod
    def declare_outputs(cls):
        """A class method to declare all the output variables provided by
        this interface.

        Returns:
          list: List of output identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example:

              outputs = ["My:first:variable",
                         "My:third:variable",
                        ]
        """

        outputs = None

        return outputs

    @classmethod
    def declare_optional(cls):
        """A class method to declare all the variables which should be flagged
        as optional.

        Returns:
          list: List of optional variable identifiers

        Note:
          Currently only inputs marked as optional have any logical effect.
          However, this may change in future releases hence the general
          approach.

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the declare_inputs output. For example:

              optional = ["My:first:variable",
                         ]
        """

        option_list = [
            "site.corridor_boundary",
            "site.lease_boundary",
            "corridor.landing_point",
        ]

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example:

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "projection": "site.projection",
            "corridor_poly": "site.corridor_boundary",
            "lease_poly": "site.lease_boundary",
            "landing_point": "corridor.landing_point",
        }

        return id_map

    def connect(self):
        """The connect method is used to execute the external program and
        populate the interface data store with values.

        Note:
          Collecting data from the interface for use in the external program
          can be accessed using self.data.my_input_variable. To put new values
          into the interface once the program has run we set
          self.data.my_output_variable = value

        """
        pass



#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import shutil
import tempfile
from copy import deepcopy
from pathlib import Path
from typing import Optional, Union

import matplotlib.pyplot as plt
from mdo_engine.boundary.interface import (
    AutoInterface,
    Interface,
    MetaInterface,
    QueryInterface,
    RawInterface,
)
from mdo_engine.control.data import DataStorage, DataValidation
from mdo_engine.control.factory import InterfaceFactory
from mdo_engine.control.pipeline import Sequencer
from mdo_engine.control.simulation import Controller, Loader
from mdo_engine.control.sockets import NamedSocket, Socket
from mdo_engine.entity.data import DataCatalog, DataPool
from mdo_engine.entity.simulation import Simulation
from mdo_engine.utilities.data import check_integrity
from mdo_engine.utilities.misc import OrderedSet

import dtocean_plugins.core as core_interfaces
import dtocean_plugins.modules as module_interfaces
import dtocean_plugins.plots as plot_interfaces
import dtocean_plugins.themes as theme_interfaces
from dtocean_plugins.core import FileInputInterface, FileOutputInterface
from dtocean_plugins.plots.base import PlotInterface

from . import data as core_data
from .utils.database import MIN_DB_VERSION, get_database
from .utils.files import package_dir, unpack_archive

StrOrPath = Union[str, Path]

MODULE_LOGGER = logging.getLogger(__name__)
INTERFACE_MODULES = [
    core_interfaces,
    module_interfaces,
    plot_interfaces,
    theme_interfaces,
]


class AutoRaw(AutoInterface, RawInterface):
    def __init__(self):
        AutoInterface.__init__(self)
        RawInterface.__init__(self)

    @classmethod
    def get_connect_name(cls):
        return None


class AutoQuery(AutoInterface, QueryInterface):
    def __init__(self):
        AutoInterface.__init__(self)
        QueryInterface.__init__(self)

    @classmethod
    def get_connect_name(cls):
        return "auto_db"


class AutoPlot(AutoInterface, PlotInterface):
    def __init__(self):
        AutoInterface.__init__(self)
        PlotInterface.__init__(self)

    @classmethod
    def get_connect_name(cls):
        return "auto_plot"


class AutoFileInput(AutoInterface, FileInputInterface):
    def __init__(self):
        AutoInterface.__init__(self)
        FileInputInterface.__init__(self)

    @classmethod
    def get_connect_name(cls):
        return "auto_file_input"

    @classmethod
    def get_method_names(cls):
        return ["get_valid_extensions"]


class AutoFileOutput(AutoInterface, FileOutputInterface):
    def __init__(self):
        AutoInterface.__init__(self)
        FileOutputInterface.__init__(self)

    @classmethod
    def get_connect_name(cls):
        return "auto_file_output"

    @classmethod
    def get_method_names(cls):
        return ["get_valid_extensions"]


class OrderedSim(Simulation):
    """The main class is the simulation which holds all of the information
    about the system."""

    _hub_defs = [
        {
            "name": "project",
            "interface": "ProjectInterface",
            "type": "Hub",
            "force unavailable": None,
            "no complete": False,
        },
        {
            "name": "modules",
            "interface": "ModuleInterface",
            "type": "Pipeline",
            "force unavailable": None,
            "no complete": False,
        },
        {
            "name": "themes",
            "interface": "ThemeInterface",
            "type": "Hub",
            "force unavailable": ("modules",),
            "no complete": True,
        },
    ]

    def __init__(self, title=None):
        super(OrderedSim, self).__init__(title)
        self.level_map = {}
        self.output_scope = None
        self._execution_level = None
        self._inspection_level = None
        self._hub_order = []
        self._hub_input_status = None
        self._hub_output_status = None
        self._hub_queue = None
        self._force_unvailable = None

        # Functional initialisation
        self._hub_queue = self._init_hub_queue()

    def _init_hub_queue(self):
        result = deepcopy(self._hub_defs)

        return result

    def set_unavailable_variables(self, variable_names=None):
        self._force_unvailable = variable_names

    def get_unavailable_variables(self):
        return self._force_unvailable

    def set_inspection_level(self, level):
        MODULE_LOGGER.debug("Setting inspection level to {}".format(level))

        self._inspection_level = level

    def get_inspection_level(self):
        return self._inspection_level

    def set_execution_level(self, level):
        MODULE_LOGGER.debug("Setting execution level to {}".format(level))

        self._execution_level = level

    def get_execution_level(self):
        return self._execution_level

    def next_hub_definition(self):
        if not self._hub_queue:
            errStr = "No Hubs are queued for creation."
            raise RuntimeError(errStr)

        return self._hub_queue.pop(0)

    def set_hub(self, hub_id, hub):
        self._hubs[hub_id] = hub
        self._hub_order.append(hub_id)

    def get_hub_order(self):
        if not self._hub_order:
            result = None
        else:
            result = self._hub_order

        return result

    def set_input_status(self, hub_input_status):
        self._hub_input_status = hub_input_status

    def set_output_status(self, hub_output_status):
        self._hub_output_status = hub_output_status

    def get_input_ids(
        self, hub_id=None, interface_name=None, valid_statuses=None
    ):
        result = self._get_ids(
            self._hub_input_status, hub_id, interface_name, valid_statuses
        )

        return result

    def get_output_ids(
        self, hub_id=None, interface_name=None, valid_statuses=None
    ):
        result = self._get_ids(
            self._hub_output_status, hub_id, interface_name, valid_statuses
        )

        return result

    def get_input_status(self, hub_id, interface_name) -> Optional[dict]:
        if self._hub_input_status is None:
            return None

        return self._hub_input_status[hub_id][interface_name]

    def get_output_status(self, hub_id, interface_name):
        if self._hub_output_status is None:
            return None

        return self._hub_output_status[hub_id][interface_name]

    def stamp(self, simulation):
        simulation.level_map = deepcopy(self.level_map)
        simulation._execution_level = deepcopy(self._execution_level)
        simulation._inspection_level = deepcopy(self._inspection_level)
        simulation._hub_order = deepcopy(self._hub_order)
        simulation._hub_input_status = deepcopy(self._hub_input_status)
        simulation._hub_output_status = deepcopy(self._hub_output_status)
        simulation._hub_queue = deepcopy(self._hub_queue)
        simulation._force_unvailable = deepcopy(self._force_unvailable)

        return simulation

    def _get_ids(
        self, sim_status, hub_id=None, interface_name=None, valid_statuses=None
    ):
        if sim_status is None:
            return []

        if hub_id is None:
            hub_ids = sim_status.keys()
        else:
            hub_ids = [hub_id]

        all_ids = []

        for this_hub_id in hub_ids:
            hub_status = sim_status[this_hub_id]

            if interface_name is None:
                interface_names = hub_status.keys()
            else:
                interface_names = [interface_name]

            for this_interface_name in interface_names:
                input_status = hub_status[this_interface_name]

                if valid_statuses is not None:
                    all_ids += [
                        k
                        for k, v in input_status.items()
                        if any(x == v for x in valid_statuses)
                    ]

                else:
                    all_ids += input_status.keys()

        return list(set(all_ids))


class Project:
    """Class to store simulations, data pool and other project related data."""

    def __init__(self, title: str, pool: Optional[DataPool] = None):
        self.title = title
        self._simulations: list[OrderedSim] = []
        self._active_index: Optional[int] = None
        self._db_cred: Optional[dict[str, str]] = None

        if pool is None:
            self._pool = DataPool()
        else:
            self._pool = pool

    def is_active(self):
        result = False
        if self._active_index is not None:
            result = True

        return result

    def get_active_index(self):
        return self._active_index

    def set_active_index(self, index=None, title=None):
        if index is None and title is None:
            errStr = "Either an index or simulation title is required"
            raise ValueError(errStr)

        if index is None:
            index = self._get_index(title)

        n_sims = len(self._simulations)

        if index is None or index > n_sims or index < 0:
            name = index
            if title:
                name = title

            errStr = "Simulation {} not found".format(name)
            raise ValueError(errStr)

        if index == self._active_index:
            return

        self._set_active_index(index)

    def add_simulation(self, simulation, set_active=False):
        index = self._set_simulation(simulation)

        if self._active_index is None or set_active:
            self._set_active_index(index)

    def remove_simulation(self, index=None, title=None, active_index=None):
        if index is None and title is None:
            errStr = "Either an index or simulation title is required"
            raise ValueError(errStr)

        if index is None:
            index = self._get_index(title)
            assert index is not None

        if active_index is None:
            active_index = 0

        simulation = self._simulations.pop(index)
        self._set_active_index(active_index)

        return simulation

    def get_simulation_indexes(self, titles, raise_if_missing=True):
        sim_indexes = [self._get_index(x, raise_if_missing) for x in titles]

        return sim_indexes

    def get_simulation_titles(self, indexes=None):
        if not self.is_active():
            return None

        if indexes is None:
            sim_titles = [x.get_title() for x in self._simulations]
        else:
            sim_titles = [self.get_simulation_title(index=x) for x in indexes]

        sim_titles = [
            title if title is not None else "" for title in sim_titles
        ]

        if not sim_titles:
            sim_titles = None

        return sim_titles

    def get_simulation_title(self, index=None, title=None):
        if not self.is_active():
            return None

        simulation = self._get_simulation(index, title)
        sim_title = simulation.get_title()

        return sim_title

    def set_simulation_title(self, new_title, index=None, title=None):
        simulation = self._get_simulation(index, title)

        if title is None:
            title = simulation.get_title()

        # Skip matching simulation titles
        if index is None and title is not None and title == new_title:
            return

        # Check if the title is unique. None removes title
        if new_title is not None:
            all_titles = self.get_simulation_titles()

            if all_titles is not None and new_title in all_titles:
                errStr = ("Simulation with title {} already " "exists").format(
                    new_title
                )
                raise ValueError(errStr)

        simulation.set_title(new_title)

        self._set_simulation(simulation)

    def get_pool(self):
        return self._pool

    def get_simulation(self, index=None, title=None):
        return self._get_simulation(index, title)

    def set_database_credentials(self, database_cred):
        self._db_cred = database_cred

    def get_database_credentials(self):
        return self._db_cred

    def check_integrity(self):
        result = check_integrity(self._pool, self._simulations)

        return result

    def to_project(self):
        new_project = Project(self.title)

        new_project._pool = deepcopy(self._pool)  # pylint: disable=protected-access
        new_project._simulations = deepcopy(self._simulations)  # pylint: disable=protected-access
        new_project._active_index = self._active_index  # pylint: disable=protected-access
        new_project._db_cred = deepcopy(self._db_cred)  # pylint: disable=protected-access

        return new_project

    def _get_index(self, title, raise_if_missing=True):
        sim_index = None

        for i, simulation in enumerate(self._simulations):
            sim_title = simulation.get_title()

            if title == sim_title:
                sim_index = i
                break

        if sim_index is None and raise_if_missing:
            errStr = "Simualtion {} not found".format(title)
            raise ValueError(errStr)

        return sim_index

    def _set_active_index(self, index):
        n_sims = len(self._simulations)

        if n_sims == 0:
            self._active_index = None
            return

        if index > n_sims or index < 0:
            errStr = "Index {} is out of range".format(index)
            raise ValueError(errStr)

        self._active_index = index

    def _get_simulation(self, index=None, title=None):
        """If both index and title are none then return the active simulation."""

        if index is not None:
            n_sims = len(self._simulations)

            if index > n_sims or index < 0:
                errStr = "Simulation index out of range."
                raise ValueError(errStr)

            sim_index = index

        elif title is not None:
            sim_index = self._get_index(title)
            assert sim_index is not None

        else:
            if not self.is_active():
                errStr = ("Project {} has no active " "simulation").format(
                    self.title
                )
                raise RuntimeError(errStr)

            sim_index = self._active_index
            assert sim_index is not None

        return self._simulations[sim_index]

    def _set_simulation(self, simulation, index=None):
        n_sims = len(self._simulations)

        if index is not None and (index > n_sims or index < 0):
            errStr = "Simulation index out of range."
            raise ValueError(errStr)

        if index is None:
            # Check for a matching title
            title = simulation.get_title()

            for i, sim in enumerate(self._simulations):
                sim_title = sim.get_title()

                if title == sim_title:
                    index = i
                    break

        if index is None:
            self._simulations.append(simulation)
            index = len(self._simulations) - 1
        else:
            self._simulations[index] = simulation

        return index

    def __len__(self):
        return len(self._simulations)


class Core:
    """Class to initiate and manipulate projects."""

    # Socket configuration constants
    _hub_sockets = ("ProjectInterface", "ModuleInterface", "ThemeInterface")

    _ext_sockets = (
        "FileInputInterface",
        "FileOutputInterface",
        "QueryInterface",
        "RawInterface",
        "PlotInterface",
    )

    _auto_classes = (
        AutoFileInput,
        AutoFileOutput,
        AutoPlot,
        AutoRaw,
        AutoQuery,
    )

    _markers = {
        "initial": "initial",
        "register": "start",
        "output": "output",
        "local": "local",
        "global": "global",
    }

    def __init__(self):
        self.data_catalog = self._create_data_catalog()
        self.loader, self.control = self._create_control()
        self.socket_map = self._create_sockets()

        # Set up plotting (must happen after sockets are created)
        self._init_plots()

    def _create_control(self):
        data_store = DataStorage(core_data)
        sequencer = Sequencer(
            self._hub_sockets, INTERFACE_MODULES, warn_import=True
        )

        loader = Loader(data_store)
        control = Controller(data_store, sequencer)

        return loader, control

    def _create_data_catalog(self):
        catalog = DataCatalog()
        validation = DataValidation(core_data.CoreMetaData)
        validation.update_data_catalog_from_definitions(catalog, core_data)

        return catalog

    def _create_sockets(self) -> dict[str, Socket]:
        socket_dict = {}

        for socket_str in self._ext_sockets:
            socket = self._build_named_socket(socket_str)
            socket_dict[socket_str] = socket

        for auto_cls in self._auto_classes:
            socket = self._build_auto_socket(auto_cls)
            socket_dict[auto_cls.__name__] = socket

        return socket_dict

    def _init_plots(self):
        """Set matplotlib to non-interactive (in case it is set to interactive
        in an imported module)"""

        plt.ioff()

    def new_project(self, project_title, simulation_title="Default"):
        new_project = Project(project_title)
        self.new_simulation(new_project, simulation_title)

        return new_project

    def dump_project(self, project, dump_path: StrOrPath):
        dump_path = Path(dump_path)

        # A data store is required
        data_store = DataStorage(core_data)

        dtop_dir_path = tempfile.mkdtemp()

        # Copy the project before editing and ensure type Project
        project_copy = project.to_project()

        # Serialise the pool
        pool_dir = os.path.join(dtop_dir_path, "pool")

        if os.path.exists(pool_dir):
            shutil.rmtree(pool_dir)
        os.makedirs(pool_dir)

        pool = project_copy.get_pool()
        serial_pool = data_store.serialise_pool(
            pool,
            pool_dir,
            root_dir=dtop_dir_path,
        )

        # Now iterate through the simulations
        sim_dicts = []

        for i, simulation in enumerate(project_copy._simulations):
            sim_dir_name = "simulation_{}".format(i)
            sim_dir = os.path.join(dtop_dir_path, sim_dir_name)

            if os.path.exists(sim_dir):
                shutil.rmtree(sim_dir)
            os.makedirs(sim_dir)

            sim_file_name = "{}.json".format(sim_dir_name)
            sim_file_path = os.path.join(sim_dir, sim_file_name)

            serial_sim = self.control.serialise_simulation(
                simulation,
                sim_dir,
                root_dir=dtop_dir_path,
            )

            # Add extra keys for OrderedSim
            serial_sim["level_map"] = simulation.level_map
            serial_sim["output_scope"] = simulation.output_scope
            serial_sim["execution_level"] = simulation._execution_level
            serial_sim["inspection_level"] = simulation._inspection_level
            serial_sim["hub_order"] = simulation._hub_order
            serial_sim["hub_input_status"] = simulation._hub_input_status
            serial_sim["hub_output_status"] = simulation._hub_output_status
            serial_sim["force_unvailable"] = simulation._force_unvailable

            with open(sim_file_path, "w") as fstream:
                json.dump(serial_sim, fstream)

            sim_store_path = Path(sim_dir_name) / sim_file_name
            sim_dicts.append({"file_path": sim_store_path.as_posix()})

        # Create a serial representation for the project and save it
        serial_project = {
            "version": 1,
            "title": project.title,
            "pool": serial_pool,
            "simulations": sim_dicts,
            "active_index": project._active_index,
        }
        project_file_path = os.path.join(dtop_dir_path, "project.json")

        with open(project_file_path, "w") as fstream:
            json.dump(serial_project, fstream)

        # OK need to consider if we want a dtop file or a directory first.
        if dump_path.suffix == ".dtop":
            archive = True
        elif dump_path.is_dir():
            archive = False
        else:
            errStr = (
                "Argument dump_path must either be an existing "
                "directory or a file path with .dtop extension"
            )
            raise ValueError(errStr)

        # Package the directory
        package_dir(dtop_dir_path, dump_path, archive)

    def load_project(self, load_path):
        # A data store is required
        data_store = DataStorage(core_data)

        # Flag to remove project directory
        remove_dtop_dir = False

        # OK need to consider if we have a dtop file or a directory first.
        # If its a dtop file them unzip it.
        if os.path.isfile(load_path) and ".dtop" in load_path:
            # Unzip the file to a temporary directory
            dtop_dir_path = tempfile.mkdtemp()
            unpack_archive(load_path, dtop_dir_path)
            remove_dtop_dir = True

        elif os.path.isdir(load_path):
            dtop_dir_path = load_path

        else:
            errStr = (
                "Argument load_path must either be a directory or a "
                "file with .dtop extension"
            )
            raise ValueError(errStr)

        # Now load the project
        project_file_path = os.path.join(dtop_dir_path, "project.json")

        with open(project_file_path, "r") as fstream:
            serial_project = json.load(fstream)

        if serial_project["version"] != 1:
            raise RuntimeError("Data version not recognised")

        # Now iterate through the serial boxes
        simulations = []

        for sim_dict in serial_project["simulations"]:
            sim_file_relative = Path(sim_dict["file_path"])
            sim_file_path = Path(dtop_dir_path) / sim_file_relative

            with open(sim_file_path, "r") as fstream:
                serial_sim = json.load(fstream)

            simulation = self.control.deserialise_simulation(
                serial_sim,
                root_dir=dtop_dir_path,
                sim_class=OrderedSim,
            )
            assert isinstance(simulation, OrderedSim)

            simulation.level_map = serial_sim["level_map"]
            simulation.output_scope = serial_sim["output_scope"]
            simulation._execution_level = serial_sim["execution_level"]
            simulation._inspection_level = serial_sim["inspection_level"]
            simulation._hub_order = serial_sim["hub_order"]
            simulation._hub_input_status = serial_sim["hub_input_status"]
            simulation._hub_output_status = serial_sim["hub_output_status"]
            simulation._force_unvailable = serial_sim["force_unvailable"]

            simulations.append(simulation)

        # Deserialise the pool
        pool = data_store.deserialise_pool(
            serial_project["pool"],
            self.data_catalog,
            root_dir=dtop_dir_path,
            warn_missing=True,
            warn_load=True,
        )

        load_project = Project(serial_project["title"], pool)
        load_project._simulations = simulations
        load_project._active_index = serial_project["active_index"]

        # Remove the project directory if necessary
        if remove_dtop_dir:
            shutil.rmtree(dtop_dir_path)

        # Reset the input / output statuses
        for simulation in simulations:
            self.set_interface_status(load_project, simulation)

        return load_project

    def new_simulation(self, project, title=None):
        # If given, check if the title is unique
        if title is not None:
            all_titles = project.get_simulation_titles()

            if all_titles is not None and title in all_titles:
                errStr = ("Simulation with title {} already " "exists").format(
                    title
                )
                raise ValueError(errStr)

        new_sim = OrderedSim(title)
        new_sim.set_execution_level(self._markers["initial"])
        new_sim.set_inspection_level(self._markers["initial"])
        project.add_simulation(new_sim)

        self.register_level(project, self._markers["initial"], None)

    def clone_simulation(
        self,
        project,
        title=None,
        sim_index=None,
        sim_title=None,
        set_active=True,
    ):
        pool = project.get_pool()
        simulation = project.get_simulation(sim_index, sim_title)

        force_title = None
        null_title = False

        if title is None:
            null_title = True
        else:
            force_title = title

        simulation_clone = self.control.copy_simulation(
            pool, simulation, force_title, null_title
        )

        project.add_simulation(simulation_clone, set_active)

    def import_simulation(
        self,
        src_project,
        dst_project,
        dst_sim_title,
        src_sim_index=None,
        src_sim_title=None,
        set_active=True,
    ):
        src_pool = src_project.get_pool()
        dst_pool = dst_project.get_pool()

        src_simulation = src_project.get_simulation(
            src_sim_index, src_sim_title
        )

        dst_simulation = self.control.import_simulation(
            src_pool, dst_pool, src_simulation, dst_sim_title
        )

        dst_project.add_simulation(dst_simulation, set_active)

    def remove_simulation(
        self, project, sim_index=None, sim_title=None, active_index=None
    ):
        pool = project.get_pool()
        simulation = project.remove_simulation(
            sim_index, sim_title, active_index
        )

        self.control.remove_simulation(pool, simulation)

    def new_hub(self, project):
        # For DTOcean the hubs are assumed to come one after another, but this
        # is not a requirement of mdo_engine. To facilitate this we consume the
        # _hub_queue attribute of the Project object
        simulation = project.get_simulation()

        hub_definition = simulation.next_hub_definition()

        if hub_definition["type"] == "Hub":
            self.control.create_new_hub(
                simulation,
                hub_definition["interface"],
                hub_definition["name"],
                hub_definition["no complete"],
            )

        elif hub_definition["type"] == "Pipeline":
            self.control.create_new_pipeline(
                simulation,
                hub_definition["interface"],
                hub_definition["name"],
                hub_definition["no complete"],
            )

        else:
            raise ValueError

    def get_metadata(self, identifier):
        self.check_valid_variable(identifier)
        metadata = self.data_catalog.get_metadata(identifier)

        return metadata

    def is_valid_variable(self, identifier):
        all_vars = self.data_catalog.get_variable_identifiers()

        result = False

        if identifier in all_vars:
            result = True

        return result

    def check_valid_variable(self, identifier):
        if not self.is_valid_variable(identifier):
            errStr = (
                "Variable with identifier {} is not contained in the "
                "data catalogue"
            ).format(identifier)
            raise ValueError(errStr)

    def has_data(self, project, identifier):
        self.check_valid_variable(identifier)
        simulation = project.get_simulation()

        return self.control.has_data(simulation, identifier)

    def get_data_value(self, project, identifier):
        self.check_valid_variable(identifier)
        pool = project.get_pool()
        simulation = project.get_simulation()

        if not self.has_data(project, identifier):
            errStr = (
                'Variable ID "{}" is not contained in the active data ' "state."
            ).format(identifier)
            raise ValueError(errStr)

        data_value = self.control.get_data_value(pool, simulation, identifier)

        return data_value

    def add_datastate(
        self,
        project,
        level=None,
        identifiers=None,
        values=None,
        update_status=True,
        use_objects=False,
        log_exceptions=False,
    ):
        pool = project.get_pool()
        simulation = project.get_simulation()

        self.control.add_datastate(
            pool,
            simulation,
            level,
            self.data_catalog,
            identifiers,
            values,
            use_objects=use_objects,
            log_exceptions=log_exceptions,
        )

        # Log the identifiers added
        if level is None:
            msg_str = "Data added for "
        else:
            msg_str = "Data added with level '{}' for ".format(level)

        if identifiers is not None:
            # Filter data with non-None values
            log_identifiers = []

            assert values is not None
            for identifier, value in zip(identifiers, values):
                if value is not None:
                    log_identifiers.append(identifier)

            if len(log_identifiers) == 1:
                msg_str += "identifier '{}'".format(log_identifiers[0])

            elif len(log_identifiers) > 1:
                id_tabs = ["    {}".format(x) for x in log_identifiers]
                id_str = "\n".join(id_tabs)
                msg_str += "identifiers:\n{}".format(id_str)

            if log_identifiers:
                MODULE_LOGGER.info(msg_str)

        if not update_status:
            return

        self.set_interface_status(project)

    def mask_states(
        self,
        project,
        simulation=None,
        search_str=None,
        mask_after=None,
        no_merge=False,
        update_status=True,
    ):
        if simulation is None:
            simulation = project.get_simulation()

        # Mask all output states after the given level
        n_masks = self.control.mask_states(
            simulation, search_str, mask_after, no_merge
        )

        if not update_status or n_masks == 0:
            return

        self.set_interface_status(project)

    def unmask_states(
        self,
        project,
        simulation=None,
        search_str=None,
        no_merge=False,
        update_status=True,
    ):
        if simulation is None:
            simulation = project.get_simulation()

        # Remove all existing masks
        n_unmasks = self.control.unmask_states(simulation, search_str, no_merge)

        if not update_status or n_unmasks == 0:
            return

        self.set_interface_status(project)

    def dump_datastate(self, project, dump_path, mask=None):
        data_store = DataStorage(core_data)

        def get_subsets():
            pool = project.get_pool()
            simulation = project.get_simulation()

            merged_state = self.loader.create_merged_state(simulation)
            save_pool, save_state = data_store.create_pool_subset(
                pool, merged_state
            )

            return save_pool, save_state

        # Allow a mask to applied before dumping
        if mask is not None:
            self.mask_states(project, search_str=mask)

        # Get the pool and datastate subsets
        save_pool, save_state = get_subsets()

        if mask is not None:
            self.unmask_states(project)

        # Serialise the pool
        dtos_dir_path = tempfile.mkdtemp()
        pool_dir = os.path.join(dtos_dir_path, "pool")

        if os.path.exists(pool_dir):
            shutil.rmtree(pool_dir)
        os.makedirs(pool_dir)

        serial_pool = data_store.serialise_pool(
            save_pool,
            pool_dir,
            root_dir=dtos_dir_path,
        )
        # Now dump the pool
        pool_file_path = os.path.join(dtos_dir_path, "pool.json")

        with open(pool_file_path, "w") as fstream:
            json.dump(serial_pool, fstream)

        # Serialise the datastate
        file_path = os.path.join(dtos_dir_path, "datastate_dump.json")
        state_dict = save_state.dump()

        with open(file_path, "w") as json_file:
            json.dump(state_dict, json_file)

        # OK need to consider if we want a file or a directory first.
        errStr = (
            "Argument dump_path must either be an existing "
            "directory or a file path with .dtos extension"
        )

        if os.path.splitext(dump_path)[1] == ".dtos":
            archive = True
        elif os.path.isdir(dump_path):
            archive = False
        else:
            raise ValueError(errStr)

        # Package the directory
        package_dir(dtos_dir_path, dump_path, archive)

    def load_datastate(self, project, load_path, exclude=None, overwrite=True):
        # A data store is required
        data_store = DataStorage(core_data)

        # Get the input ids in the active simulation
        simulation = project.get_simulation()

        # Avoid filling unavailable or overwritten inputs
        valid_statuses = ["required", "optional"]
        if overwrite:
            valid_statuses.append("satisfied")

        input_ids = simulation.get_input_ids(valid_statuses=valid_statuses)

        # Flag to remove datastate directory
        remove_dtos_dir = False

        # OK need to consider if we have a dtos file or a directory first.
        # If its a dtos file them unzip it.
        if os.path.isfile(load_path) and ".dtos" in load_path:
            # Unzip the file to a temporary directory
            dtos_dir_path = tempfile.mkdtemp()
            unpack_archive(load_path, dtos_dir_path)
            remove_dtos_dir = True

        elif os.path.isdir(load_path):
            dtos_dir_path = load_path

        else:
            errStr = (
                "Argument load_path must either be a directory or a file"
                "with .dtos extension"
            )
            raise ValueError(errStr)

        # Load datastate json
        load_path = os.path.join(dtos_dir_path, "datastate_dump.json")

        with open(load_path, "r") as json_file:
            dump_dict = json.load(json_file)

        state_data: dict = dump_dict["data"]

        # Now load the pool data
        pool_file_path = os.path.join(dtos_dir_path, "pool.json")

        with open(pool_file_path, "r") as fstream:
            serial_pool = json.load(fstream)

        # Deserialise the pool
        temp_pool = data_store.deserialise_pool(
            serial_pool,
            self.data_catalog,
            root_dir=dtos_dir_path,
            warn_missing=True,
            warn_load=True,
        )

        # Remove the project directory if necessary
        if remove_dtos_dir:
            shutil.rmtree(dtos_dir_path)

        # Create a new datastate in the existing pool with the loaded data
        var_ids = []
        var_objs = []

        for var_id, data_index in state_data.items():
            if not self.is_valid_variable(var_id):
                msgStr = (
                    'Variable ID "{}" is not contained in the data ' "catalog"
                ).format(var_id)
                MODULE_LOGGER.warning(msgStr)

                continue

            if var_id not in input_ids:
                msgStr = (
                    'Variable ID "{}" is not an input to the current '
                    "simulation"
                ).format(var_id)
                MODULE_LOGGER.info(msgStr)

                continue

            if exclude is not None and exclude in var_id:
                continue

            data_obj = temp_pool.get(data_index)

            var_ids.append(var_id)
            var_objs.append(data_obj)

        self.add_datastate(
            project,
            identifiers=var_ids,
            values=var_objs,
            use_objects=True,
        )

    def get_levels(
        self,
        project,
        show_masked=True,
        sim_index=None,
        sim_title=None,
    ):
        simulation = project.get_simulation(sim_index, sim_title)
        levels = OrderedSet(
            simulation.get_active_levels(show_masked=show_masked)
        )

        return levels

    def get_level_values(
        self,
        project,
        data_identity,
        levels=None,
        force_masks=None,
        sim_index=None,
        sim_title=None,
    ):
        pool = project.get_pool()
        simulation = project.get_simulation(sim_index, sim_title)

        level_results = self.control.get_level_values(
            pool,
            simulation,
            data_identity,
            levels=levels,
            force_masks=force_masks,
        )

        return level_results

    def get_project_values(
        self,
        project,
        data_identity,
        level,
        force_indexes=None,
        allow_none=False,
    ):
        """Collect the value of a given identity at a given level for all
        simulations in the project"""

        pool = project.get_pool()
        project_values = []

        if force_indexes is None:
            sim_indexes = range(len(project))
        else:
            sim_indexes = force_indexes

        for i in sim_indexes:
            simulation = project.get_simulation(index=i)
            sim_title = simulation.get_title()

            sim_value = self.control.get_data_value(
                pool,
                simulation,
                data_identity,
                level=level,
                check_identity=True,
            )

            if sim_value is None and not allow_none:
                continue

            project_values.append((sim_title, sim_value))

        if not project_values:
            project_values = None

        return project_values

    def register_level(self, project, level, interface_name):
        pool = project.get_pool()
        simulation = project.get_simulation()

        self.control.add_datastate(pool, simulation, level)

        simulation.level_map[level] = interface_name

    def inspect_level(
        self,
        project,
        level,
        inspection_level=None,
        update_status=True,
        skip_missing=False,
        force=False,
    ):
        """Inspect the data at a level, masking all output levels after
        it"""

        simulation = project.get_simulation()

        # Sanitise the level
        level = level.lower()

        if not force and level == simulation.get_inspection_level():
            return

        # Check if state is in the active levels
        if level not in self.get_levels(project):
            active_level_str = ", ".join(self.get_levels(project))
            errStr = (
                "Level {} is not available. Active levels are " "{}"
            ).format(level, active_level_str)
            raise ValueError(errStr)

        if inspection_level is None:
            inspection_level = level

        if inspection_level not in simulation.level_map:
            level_str = ", ".join(simulation.level_map.keys())
            msgStr = (
                "System level '{}' not valid for inspection. Must be "
                "one of {}"
            ).format(level, level_str)

            if skip_missing:
                msgStr += ". Skipping"
                MODULE_LOGGER.debug(msgStr)
                return

            raise RuntimeError(msgStr)

        MODULE_LOGGER.info("Inspecting level {}".format(level))

        # Remove all existing masks
        self.unmask_states(project, no_merge=True, update_status=False)

        # Mask all output states after the given level
        self.mask_states(
            project,
            search_str=self._markers["output"],
            mask_after=level,
            update_status=False,
        )

        simulation.set_inspection_level(inspection_level)

        if not update_status:
            return

        self.set_interface_status(project)

    def reset_level(
        self,
        project,
        level=None,
        preserve_level=False,
        force_scheduled=None,
        skip_missing=False,
    ):
        """Prepare the simulation for re-execution at the given level"""

        pool = project.get_pool()
        simulation = project.get_simulation()

        if level == simulation.get_execution_level():
            return

        if level is None:
            level = simulation.get_inspection_level()

        else:
            level = level.lower()

            # Bring the hubs to the correct position if the inspection level
            # has been changed. Inspection level must match a module name
            # unless skip_missing flag is True
            if level not in simulation.level_map:
                level_str = ", ".join(simulation.level_map.keys())
                msgStr = (
                    "System level '{}' not valid for execution. Must be "
                    "one of {}"
                ).format(level, level_str)

                if skip_missing:
                    msgStr += ". Skipping"
                    MODULE_LOGGER.debug(msgStr)
                    return

                raise RuntimeError(msgStr)

        self.inspect_level(project, level, update_status=False, force=True)

        MODULE_LOGGER.info("Resetting to level {}".format(level))

        # Mask all register states after the given level
        self.mask_states(
            project,
            search_str=self._markers["register"],
            mask_after=level,
            no_merge=True,
            update_status=False,
        )

        # Mask the given level
        if not preserve_level:
            self.mask_states(
                project, search_str=level, no_merge=True, update_status=False
            )

        interface_name = simulation.level_map[level]
        self._schedule_interface(project, interface_name)

        # Later interfaces can now not be scheduled so delete the masked
        # states.
        self.control.delete_masked_states(pool, simulation)

        simulation.set_execution_level(level)

        # Switch the force available flag on a given hub
        if force_scheduled is not None:
            hub = simulation.get_hub(force_scheduled)
            hub.force_completed = False

        self.set_interface_status(project)

    def set_interface_status(self, project, simulation=None):
        # This order is important as the input status relies on the output
        # status
        self._set_outputs_status(project, simulation)
        self._set_inputs_status(project, simulation)

    def can_load_interface(self, project: Project, interface, check_id=None):
        pool = project.get_pool()
        simulation = project.get_simulation()

        if check_id:
            if not self.loader.input_available(
                pool, simulation, interface, check_id
            ):
                return False

        result = self.loader.can_load(pool, simulation, interface)

        return result

    def load_interface(self, project, interface, skip_vars=None):
        pool = project.get_pool()
        simulation = project.get_simulation()

        interface = self.loader.load_interface(
            pool, simulation, interface, skip_vars
        )

        return interface

    def connect_interface(self, project: Project, interface: Interface):
        # TODO: Is this pre-population something to do in mdo_engine?
        # If its a QueryInterface try to connect the database
        if (
            isinstance(interface, QueryInterface)
            and project.get_database_credentials() is not None
        ):
            credentials = project.get_database_credentials()
            database = get_database(
                credentials,
                timeout=60,
                echo=False,
                min_version=MIN_DB_VERSION,
            )
            interface.put_database(database)

        # If its a MetaInterface try to add meta data
        if isinstance(interface, MetaInterface):
            input_vars, _ = interface.get_inputs(True)
            output_vars = interface.get_outputs()
            all_vars = set(input_vars) | set(output_vars)

            for var in all_vars:
                meta = self.get_metadata(var)
                interface.put_meta(var, meta)

        if isinstance(interface, QueryInterface):
            interface.safe_connect()
            interface._db = None

        else:
            interface.connect()

        return interface

    def _build_named_socket(self, socket_str):
        socket = NamedSocket(socket_str)
        for interface_module in INTERFACE_MODULES:
            socket.discover_interfaces(interface_module)

        return socket

    def _build_auto_socket(self, auto_cls):
        auto_socket = Socket()
        interface_factory = InterfaceFactory(auto_cls)

        for var_id in self.data_catalog.get_variable_identifiers():
            metadata = self.data_catalog.get_metadata(var_id)

            # Catch errors here for more informative reporting:
            try:
                data_obj = self.control.get_structure(metadata.structure)
            except KeyError:
                errStr = ("Structure {} not found for variable " "{}").format(
                    metadata.structure, var_id
                )
                raise ValueError(errStr)

            # Allow auto classes with no method to pass
            if (
                auto_cls.get_connect_name() is not None
                and not interface_factory.has_connect_method(data_obj)
            ):
                continue

            AutoCls = interface_factory(metadata, data_obj)

            auto_socket.add_interface(AutoCls)

        return auto_socket

    def _get_force_unavailable(self, simulation):
        hub_force_available = {}

        for hub_definition in simulation._hub_defs:
            hub_name = hub_definition["name"]
            unavailable_hubs = hub_definition["force unavailable"]

            if unavailable_hubs is None:
                hub_force_available[hub_name] = []
            else:
                hub_force_available[hub_name] = unavailable_hubs

        return hub_force_available

    def _schedule_interface(self, project, interface_name=None):
        """Check if any of the hubs have the expected interface scheduled."""

        simulation = project.get_simulation()

        hub_ids = simulation.get_hub_order()
        other_hub_ids = hub_ids[1:]

        # If the interface name is None then reset all the connectors
        if interface_name is None:
            for hub_id in hub_ids:
                self.control.reset_hub(simulation, hub_id)

            return

        # Using a hub_id will reset all the hubs after and including the given
        # id
        elif interface_name in hub_ids:
            hub_index = hub_ids.index(interface_name)
            reset_hub_ids = hub_ids[hub_index:]

            for hub_id in reset_hub_ids:
                self.control.reset_hub(simulation, hub_id)

            return

        for hub_id in hub_ids:
            if self.control.has_interface(simulation, hub_id, interface_name):
                self.control.check_next_interface(
                    simulation, hub_id, interface_name
                )

                # It is assumed that all connectors after the rescheduled one
                # are dependent and therefore should be reset
                for other_hub_id in other_hub_ids:
                    self.control.reset_hub(simulation, other_hub_id)

                return

            else:
                other_hub_ids.pop(0)

        errStr = "Interface {} not in any Hub".format(interface_name)
        raise ValueError(errStr)

    def _set_outputs_status(self, project, simulation=None):
        if simulation is None:
            simulation = project.get_simulation()

        hub_dict = {}

        for hub_id in simulation.get_hub_order():
            if hub_id == "modules":
                inspection_interface_name = simulation.level_map[
                    simulation._inspection_level
                ]

            else:
                inspection_interface_name = None

            interface_dict = {}

            sequenced_interfaces = self.control.get_sequenced_interfaces(
                simulation, hub_id
            )

            for interface_name in sequenced_interfaces:
                outputs_status = self.control.get_output_status(
                    simulation,
                    hub_id,
                    interface_name,
                    force_last_completed=inspection_interface_name,
                )

                interface_dict[interface_name] = outputs_status

            hub_dict[hub_id] = interface_dict

        simulation.set_output_status(hub_dict)

    def _set_inputs_status(self, project, simulation=None):
        pool = project.get_pool()
        if simulation is None:
            simulation = project.get_simulation()

        hub_dict = {}
        hub_force_available = self._get_force_unavailable(simulation)

        for hub_id in simulation.get_hub_order():
            all_unavailable = []

            for hub_name in hub_force_available[hub_id]:
                hub_unavailable = self._get_unavailable_outputs(
                    project, hub_name, simulation
                )
                all_unavailable.extend(hub_unavailable)

            force_unavailable = simulation.get_unavailable_variables()

            if force_unavailable is not None:
                all_unavailable.extend(force_unavailable)

            if not all_unavailable:
                all_unavailable = None

            interface_dict = {}

            sequenced_interfaces = self.control.get_sequenced_interfaces(
                simulation, hub_id
            )

            for interface_name in sequenced_interfaces:
                inputs_status = self.control.get_input_status(
                    pool, simulation, hub_id, interface_name, all_unavailable
                )

                interface_dict[interface_name] = inputs_status

            hub_dict[hub_id] = interface_dict

        simulation.set_input_status(hub_dict)

    def _get_unavailable_outputs(self, project, hub_id, simulation=None):
        # Collect the not satisfied outputs of the modules for the input
        # status of the themes.
        if simulation is None:
            simulation = project.get_simulation()

        all_scheduled_outputs = []

        scheduled_interfaces = self.control.get_scheduled_interfaces(
            simulation, hub_id
        )

        for interface_name in scheduled_interfaces:
            out_status = simulation.get_output_status(hub_id, interface_name)

            all_scheduled_outputs.extend(out_status.keys())

        unavailable_set = set(all_scheduled_outputs)

        return list(unavailable_set)


class Connector:
    """Class for working with interface hubs."""

    def __init__(self, hub_name):
        self._hub = hub_name

    def get_force_completed(self, project):
        simulation = project.get_simulation()
        hub = simulation.get_hub(self._hub)

        return hub.force_completed

    def set_force_completed(self, core, project, value=True):
        simulation = project.get_simulation()
        hub = simulation.get_hub(self._hub)
        hub.force_completed = value

        core.set_interface_status(project)

    def get_available_interfaces(self, core, project):
        simulation = project.get_simulation()
        names = core.control.get_available_interfaces(simulation, self._hub)
        return names

    def get_active_interface_names(self, core: Core, project: Project):
        simulation = project.get_simulation()
        interface_names = core.control.get_sequenced_interfaces(
            simulation, self._hub
        )
        return interface_names

    def has_interface(self, core, project, interface_name):
        interface_names = self.get_active_interface_names(core, project)

        result = False
        if interface_name in interface_names:
            result = True

        return result

    def get_scheduled_interface_names(self, core, project):
        simulation = project.get_simulation()
        interface_names = core.control.get_scheduled_interfaces(
            simulation, self._hub
        )

        return interface_names

    def any_scheduled(self, core, project):
        interface_names = self.get_scheduled_interface_names(core, project)

        result = False
        if interface_names:
            result = True

        return result

    def get_current_interface_name(self, core, project):
        simulation = project.get_simulation()
        interface_name = core.control.get_next_interface(simulation, self._hub)

        return interface_name

    def get_completed_interface_names(self, core, project):
        simulation = project.get_simulation()
        interface_names = core.control.get_completed_interfaces(
            simulation, self._hub
        )

        return interface_names

    def get_interface_inputs(self, core, project, interface_name):
        simulation = project.get_simulation()
        pool = project.get_pool()

        interface = core.control.get_interface_obj(
            simulation, self._hub, interface_name
        )
        input_declaration, _ = interface.get_inputs()

        active_inputs = core.control._get_active_inputs(
            pool, simulation, input_declaration
        )

        return active_inputs

    def get_interface_inputs_status(self, project: Project, interface_name):
        simulation = project.get_simulation()
        inputs_status = simulation.get_input_status(self._hub, interface_name)

        return inputs_status

    def get_interface_outputs(self, core, project, interface_name):
        simulation = project.get_simulation()
        interface = core.control.get_interface_obj(
            simulation, self._hub, interface_name
        )

        output_declaration = interface.get_outputs()

        return output_declaration

    def get_interface_outputs_status(self, project, interface_name):
        simulation = project.get_simulation()
        outputs_status = simulation.get_output_status(self._hub, interface_name)

        return outputs_status

    def get_interface_weight(self, core, project, interface_name):
        """Get the weighting of the interface if set"""

        simulation = project.get_simulation()
        result = core.control.get_interface_weight(
            simulation, self._hub, interface_name
        )

        return result

    def is_interface_executable(
        self, core, project, interface_name, allow_unavailable=False
    ):
        # For the data requirments to be met all the inputs should list
        # as satisfied.
        input_status = self.get_interface_inputs_status(project, interface_name)
        if input_status is None:
            return False

        assert isinstance(input_status, dict)

        if len(input_status) == 0:
            return True

        status_values = set(input_status.values())

        result = True

        if "required" in status_values:
            result = False

        if "unavailable" in status_values and not allow_unavailable:
            result = False

        return result

    def is_auto_executable(self, core, project, allow_unavailable=False):
        # For the data requirments to be met all the inputs should list
        # as satisfied.
        scheduled_interfaces = self.get_scheduled_interface_names(core, project)

        result = True

        # Try to execute all the interfaces
        for interface_name in scheduled_interfaces:
            # Test for inputs
            can_execute = self.is_interface_executable(
                core,
                project,
                interface_name,
                allow_unavailable=allow_unavailable,
            )

            result = result & can_execute

        return result

    def is_interface_completed(self, core, project, interface_name):
        simulation = project.get_simulation()
        result = core.control.is_interface_completed(
            simulation, self._hub, interface_name
        )

        return result

    def activate_interface(
        self, core, project, interface_name, update_status=True
    ):
        simulation = project.get_simulation()

        core.control.sequence_interface(simulation, self._hub, interface_name)

        if not update_status:
            return

        core.set_interface_status(project)

    def get_interface(
        self, core, project, interface_name, allow_unavailable=False
    ):
        # Only return an interface if it can be exectuted
        if not self.is_interface_executable(
            core, project, interface_name, allow_unavailable
        ):
            errStr = (
                "Input data for interface {} can not be " "provided"
            ).format(interface_name)
            raise RuntimeError(errStr)

        simulation = project.get_simulation()
        interface = core.control.get_interface_obj(
            simulation, self._hub, interface_name
        )

        interface = self._set_inputs(core, project, interface, interface_name)

        return interface

    def execute_interface(
        self,
        core,
        project,
        interface_name,
        level=None,
        register_level=True,
        allow_unavailable=False,
        set_output_level=True,
    ):
        """Add the data required for connecting the interface and then
        connect it"""

        simulation = project.get_simulation()

        # Check the inputs
        self._test_exectuable(core, project, interface_name, allow_unavailable)

        # If level is None build a level name based on the interface name
        # and record the interface name as the current level in the core
        if level is None:
            level = interface_name.lower()
        else:
            level = level.lower()

        # Add an empty datastate at this level
        if register_level:
            registered_level = "{} {}".format(level, core._markers["register"])

            simulation.set_inspection_level(registered_level)

            core.register_level(project, registered_level, interface_name)

        # Get the inteface populated with data
        interface = self.get_interface(
            core, project, interface_name, allow_unavailable
        )

        # Execute the interface.
        interface = core.connect_interface(project, interface)

        # Get outputs from the interface
        checked_vars, checked_values = self._get_outputs(
            core, project, interface
        )

        # Mark interface completed
        core.control.set_interface_completed(
            simulation, self._hub, interface_name
        )

        # Set a level for the datastate including the output marker
        if set_output_level:
            output_level = "{} {}".format(level, core._markers["output"])
            output_level = output_level.lower()
        else:
            output_level = None

        core.add_datastate(project, output_level, checked_vars, checked_values)

        # Set the execution level to the output level
        if register_level and output_level is not None:
            simulation.set_execution_level(output_level)

    def auto_execute(
        self,
        core,
        project,
        force_level=None,
        register_level=True,
        allow_non_execution=False,
    ):
        scheduled_interfaces = self.get_scheduled_interface_names(core, project)

        # Try to execute all the interfaces
        for interface_name in scheduled_interfaces:
            # Test for inputs
            can_execute = self.is_interface_executable(
                core, project, interface_name
            )

            if not can_execute:
                if not allow_non_execution:
                    errStr = (
                        "Required inputs for interface {} are not " "satisfied."
                    ).format(interface_name)
                    raise ValueError(errStr)

                else:
                    log_msg = "Skipping interface {}".format(interface_name)
                    MODULE_LOGGER.info(log_msg)
                    continue

            log_msg = "Auto executing interface {}".format(interface_name)
            MODULE_LOGGER.info(log_msg)

            # Execute the interface
            self.execute_interface(
                core, project, interface_name, force_level, register_level
            )

    def _set_inputs(self, core, project, interface, interface_name):
        (_, optional_inputs) = interface.get_inputs()

        skip_optional = []

        if optional_inputs:
            for putvar in optional_inputs:
                # Test if the data is in the core
                if not core.has_data(project, putvar):
                    skip_optional.append(putvar)

        interface = core.load_interface(
            project, interface, skip_vars=skip_optional
        )

        return interface

    def _get_outputs(self, core, project, interface, filter_none=False):
        """Add variables from the interface into the active data state"""

        checked_vars = []
        checked_values = []

        variables = interface.get_outputs()

        # If there are no output varibles then add an empty state
        if not variables:
            return (None, None)

        raw_data = []

        for getvar in variables:
            raw_data.append(interface.get_data(getvar))

        if not filter_none:
            return variables, raw_data

        # Filter out any None values
        for var, value in zip(variables, raw_data):
            if value is not None:
                checked_vars.append(var)
                checked_values.append(value)

        return checked_vars, checked_values

    def _test_exectuable(
        self, core, project, interface_name, allow_unavailable=False
    ):
        if not self.is_interface_executable(
            core, project, interface_name, allow_unavailable
        ):
            errStr = (
                "Not all inputs of interface {} have been " "satisfied."
            ).format(interface_name)
            raise ValueError(errStr)



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2025 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The core component of the dtocean project.

.. moduleauthor:: Mathew Topper <damm_horse@yahoo.co.uk>
"""

from logging.handlers import RotatingFileHandler
from pathlib import Path

from polite_config.configuration import Logger
from polite_config.paths import ModPath, UserDataPath


def start_logging():
    """Start python logger"""

    # Pick up the configuration from the user directory if it exists
    userdir = UserDataPath("dtocean_core", "DTOcean", "config")

    appdir_path = userdir.parent
    logdir = Path(appdir_path, "logs")
    logdir.mkdir(exist_ok=True, parents=True)

    # Look for logging.yaml
    if (userdir / "logging.yaml").is_file():
        configdir = userdir
    else:
        configdir = ModPath("dtocean_core", "config")

    log = Logger(configdir)
    logger = log("dtocean_core", file_prefix=logdir)

    # Rotate any rotating file handlers
    for handler in logger.handlers:
        if isinstance(handler, RotatingFileHandler):
            handler.doRollover()

    logger.info("Begin logging for dtocean_core")



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2025 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING

from dtocean_app.utils.display import is_high_dpi
from dtocean_app.widgets.extendedcombobox import ExtendedComboBox
from mdo_engine.utilities.misc import OrderedSet
from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import Qt

from dtocean_core.pipeline import Tree
from dtocean_plugins.strategies.sensitivity import UnitSensitivity
from dtocean_plugins.strategy_guis.base import GUIStrategy, StrategyWidget

if is_high_dpi() or TYPE_CHECKING:
    from dtocean_app.designer.high.unitsensitivity import (
        Ui_UnitSensitivityWidget,
    )
else:
    from dtocean_app.designer.low.unitsensitivity import (
        Ui_UnitSensitivityWidget,
    )


class GUIUnitSensitivity(GUIStrategy, UnitSensitivity):
    """A sensitivity study on a single unit variables over a given range of
    values, adjusted before execution of a chosen module."""

    def __init__(self):
        UnitSensitivity.__init__(self)
        GUIStrategy.__init__(self)

    def allow_run(self, core, project):
        if len(project) > 1:
            return False
        return True

    def get_weight(self):
        """A method for getting the order of priority of the strategy.

        Returns:
          int
        """

        return 2

    def get_widget(self, parent, shell):
        widget = UnitSensitivityWidget(parent, shell)
        return widget


class UnitSensitivityWidget(
    QtWidgets.QWidget,
    Ui_UnitSensitivityWidget,
    StrategyWidget,
):
    config_set = QtCore.Signal()
    config_null = QtCore.Signal()
    reset = QtCore.Signal()

    def __init__(self, parent, shell):
        QtWidgets.QWidget.__init__(self, parent)
        Ui_UnitSensitivityWidget.__init__(self)
        StrategyWidget.__init__(self)

        self._var_ids = None
        self._mod_names: dict[int, str]

        self._init_ui()
        self._set_interfaces(shell)

    def _init_ui(self):
        self.setupUi(self)

        # Disble line edit
        self.lineEdit.setDisabled(True)

        # Custom boxes
        self.modBox = ExtendedComboBox(self)
        self.modBox.setObjectName("modBox")
        self.gridLayout.addWidget(self.modBox, 0, 1, 1, 1)

        self.varBox = ExtendedComboBox(self)
        self.varBox.setObjectName("varBox")
        self.gridLayout.addWidget(self.varBox, 1, 1, 1, 1)

        # Signals
        self.modBox.currentIndexChanged.connect(self._set_variables)
        self.varBox.currentIndexChanged.connect(self._line_edit_ui_switch)
        self.lineEdit.textChanged.connect(self._emit_config_signal)

    def _get_var_id(self, var_name):
        if self._var_ids is None:
            return var_name
        return self._var_ids[var_name]

    def _set_interfaces(self, shell, include_str=True):
        self._shell = shell

        self.modBox.clear()

        active_modules = shell.module_menu.get_active(shell.core, shell.project)

        self._mod_names = active_modules

        self.modBox.addItems(active_modules)
        self.modBox.setCurrentIndex(-1)

    @QtCore.Slot(int)
    def _set_variables(self, box_number):
        self.varBox.clear()

        if box_number < 0:
            return

        interface_name = self._mod_names[box_number]

        tree = Tree()

        var_id_dict = {}

        branch = tree.get_branch(
            self._shell.core, self._shell.project, interface_name
        )

        var_inputs = branch.get_inputs(self._shell.core, self._shell.project)

        unique_vars = OrderedSet(var_inputs)

        var_names = []

        for var_id in unique_vars:
            var_meta = self._shell.core.get_metadata(var_id)

            if "SimpleData" in var_meta.structure:
                if var_meta.types is None:
                    errStr = (
                        "Variable {} with SimpleData structure "
                        "requires types meta data to be "
                        "set"
                    ).format(var_id)
                    raise ValueError(errStr)

                if (
                    "int" in var_meta.types
                    or "float" in var_meta.types
                    or "str" in var_meta.types
                ):
                    title = var_meta.title

                    if var_meta.units is not None:
                        title = "{} ({})".format(title, var_meta.units[0])

                    var_names.append(title)

                if title not in var_id_dict:
                    var_id_dict[title] = var_id

        self._var_ids = var_id_dict

        self.varBox.addItems(var_names)
        self.varBox.setCurrentIndex(-1)

    @QtCore.Slot(int)
    def _line_edit_ui_switch(self, box_number):
        self.lineEdit.clear()

        if box_number != -1:
            self.lineEdit.setEnabled(True)
        else:
            self.lineEdit.setDisabled(True)

    @QtCore.Slot()
    def _emit_config_signal(self):
        line_config = str(self.lineEdit.text())

        if line_config:
            self.config_set.emit()
        else:
            self.config_null.emit()

    def get_configuration(self):
        mod_name = str(self.modBox.currentText())

        var_name = str(self.varBox.currentText())
        var_id = self._get_var_id(var_name)

        var_values = self.string2types(str(self.lineEdit.text()))

        conf_dict = {
            "module_name": mod_name,
            "variable_name": var_id,
            "variable_values": var_values,
        }

        return conf_dict

    def set_configuration(self, config_dict=None):
        if config_dict is None:
            return

        mod_name = config_dict["module_name"]
        var_id = config_dict["var_name"]
        var_values = config_dict["var_values"]

        sane_var_values = [str(x) for x in var_values]
        var_values_str = ", ".join(sane_var_values)

        index = self.modBox.findText(mod_name, Qt.MatchFlag.MatchFixedString)
        if index >= 0:
            self.modBox.setCurrentIndex(index)
        else:
            errStr = "Module {} not found in modBox".format(mod_name)
            raise ValueError(errStr)

        self._set_variables(index)

        var_meta = self._shell.core.get_metadata(var_id)
        title = var_meta.title

        if var_meta.units is not None:
            title = "{} ({})".format(title, var_meta.units[0])

        index = self.varBox.findText(title, Qt.MatchFlag.MatchFixedString)
        if index >= 0:
            self.varBox.setCurrentIndex(index)
        else:
            errStr = "Variable {} not found in varBox".format(title)
            raise ValueError(errStr)

        self.lineEdit.setText(var_values_str)



# -*- coding: utf-8 -*-

#    Copyright (C) 2021-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
import traceback
import warnings
from pathlib import Path
from typing import Union

from .. import start_logging
from ..core import Core
from ..extensions import StrategyManager
from ..menu import ModuleMenu

StrOrPath = Union[str, Path]


def warn_with_traceback(
    message, category, filename, lineno, logfile=None, line=None
):
    log = logfile if hasattr(logfile, "write") else sys.stderr
    traceback.print_stack(file=log)

    assert log is not None
    log.write(warnings.formatwarning(message, category, filename, lineno, line))


def main(
    fpath,
    save: Union[bool, StrOrPath] = True,
    full=False,
    warn=False,
    log=False,
):
    if full:
        action_str = "all scheduled modules"
    else:
        action_str = "next scheduled module"

    msg_str = "\n>>> Executing {} in project '{}'".format(action_str, fpath)
    extra_string = ""

    if warn:
        extra_string += "warning tracebacks"

    if log:
        if extra_string:
            extra_string += " and "
        extra_string += "logging"

    if extra_string:
        msg_str = "{} with {}".format(msg_str, extra_string)

    msg_str += "\n"
    print(msg_str)

    if warn:
        warnings.showwarning = warn_with_traceback

    if log:
        start_logging()

    my_core = Core()
    my_project = my_core.load_project(fpath)

    if full:
        strategy_manager = StrategyManager()
        basic_strategy = strategy_manager.get_strategy("Basic")

        start_time = time.time()
        basic_strategy.execute(my_core, my_project)

    else:
        my_menu = ModuleMenu()

        start_time = time.time()
        my_menu.execute_current(my_core, my_project)

    msg_str = "\n>>> Execution time: {}s".format(time.time() - start_time)
    print(msg_str)

    if not save:
        return

    if save is True:
        prepath, ext = os.path.splitext(fpath)
        save_path = "{}_complete{}".format(prepath, ext)
    else:
        save_path = save

    msg_str = "\n>>> Saving project to: {}".format(save_path)
    print(msg_str)

    my_core.dump_project(my_project, save_path)



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2025 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from dtocean_app.widgets.dialogs import Message
from PySide6 import QtCore

from dtocean_plugins.strategies.basic import BasicStrategy
from dtocean_plugins.strategy_guis.base import GUIStrategy, StrategyWidget


class GUIBasicStrategy(GUIStrategy, BasicStrategy):
    """A basic strategy which will run all selected modules and themes in
    sequence."""

    def __init__(self):
        BasicStrategy.__init__(self)
        GUIStrategy.__init__(self)

    def allow_run(self, core, project):
        return True

    def get_weight(self):
        """A method for getting the order of priority of the strategy.

        Returns:
          int
        """

        return 1

    def get_widget(self, parent, shell):
        widget = BasicWidget(parent, "No Configuration Required")

        return widget


class BasicWidget(Message, StrategyWidget):
    config_set = QtCore.Signal()
    config_null = QtCore.Signal()
    reset = QtCore.Signal()

    def __init__(self, parent, text):
        super(BasicWidget, self).__init__(parent, text)

    def get_configuration(self):
        return {}

    def set_configuration(self, *args):
        return

    def paintEvent(self, event):
        super(BasicWidget, self).paintEvent(event)
        self.config_set.emit()



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Created on Wed Apr 06 15:59:04 2016

.. moduleauthor:: Mathew Topper <damm_horse@yahoo.co.uk>
"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.lines import Line2D
from shapely import Point, Polygon
from shapely.plotting import patch_from_polygon

from .base import PlotInterface

BLUE = "#6699cc"
GREEN = "#32CD32"
RED = "#B20000"
GREY = "#999999"


class ArrayLeasePlot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Lease Area Array Layout"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = [
            "site.lease_boundary",
            "project.layout",
            "options.boundary_padding",
        ]

        return input_list

    @classmethod
    def declare_optional(cls):
        option_list = ["options.boundary_padding"]

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "lease_poly": "site.lease_boundary",
            "layout": "project.layout",
            "padding": "options.boundary_padding",
        }

        return id_map

    def connect(self):
        fig = plt.figure()
        ax1 = fig.add_subplot(1, 1, 1, aspect="equal")

        short_layout = {
            key.replace("device", ""): value
            for key, value in self.data.layout.items()
        }

        plot_point_dict(ax1, short_layout, "k+")
        plot_lease_boundary(ax1, self.data.lease_poly, self.data.padding)

        ax1.margins(0.1, 0.1)
        ax1.autoscale_view()

        annotate_poly(ax1, self.data.lease_poly)

        xlabel = "UTM x [$m$]"
        ylabel = "UTM y [$m$]"

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.ticklabel_format(useOffset=False)
        plt.xticks(rotation=30, ha="right")

        plt.title("Array Layout in Lease Area")
        plt.tight_layout()

        self.fig_handle = plt.gcf()


class ArrayLeasePlotNumbers(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Lease Area Array Layout (With Device Numbers)"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = [
            "site.lease_boundary",
            "project.layout",
            "options.boundary_padding",
        ]

        return input_list

    @classmethod
    def declare_optional(cls):
        option_list = ["options.boundary_padding"]

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "lease_poly": "site.lease_boundary",
            "layout": "project.layout",
            "padding": "options.boundary_padding",
        }

        return id_map

    def connect(self):
        fig = plt.figure()
        ax1 = fig.add_subplot(1, 1, 1, aspect="equal")

        short_layout = {
            key.replace("device", ""): value
            for key, value in self.data.layout.items()
        }

        plot_point_dict(ax1, short_layout, "k+", annotate=True)
        plot_lease_boundary(ax1, self.data.lease_poly, self.data.padding)

        ax1.margins(0.1, 0.1)
        ax1.autoscale_view()

        annotate_poly(ax1, self.data.lease_poly)

        xlabel = "UTM x [$m$]"
        ylabel = "UTM y [$m$]"

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.ticklabel_format(useOffset=False)
        plt.xticks(rotation=30, ha="right")

        plt.title("Array Layout in Lease Area")
        plt.tight_layout()

        self.fig_handle = plt.gcf()


class ArrayCablesPlot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Array Cable Layout"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = [
            "site.lease_boundary",
            "corridor.landing_point",
            "project.layout",
            "project.cable_routes",
            "project.substation_layout",
        ]

        return input_list

    @classmethod
    def declare_optional(cls):
        option_list = ["site.lease_boundary"]

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "lease_poly": "site.lease_boundary",
            "landing_point": "corridor.landing_point",
            "layout": "project.layout",
            "cable_routes": "project.cable_routes",
            "substation_layout": "project.substation_layout",
        }

        return id_map

    def connect(self):
        fig = plt.figure()
        ax1 = fig.add_subplot(1, 1, 1, aspect="equal")

        short_layout = {
            key.replace("device", ""): value
            for key, value in self.data.layout.items()
        }
        landing_dict = {"Export Cable Landing": self.data.landing_point}

        lplot = plot_point_dict(ax1, landing_dict, "or", "Export Cable Landing")
        dplot = plot_point_dict(ax1, short_layout, "k+", "Devices")
        splot = plot_point_dict(
            ax1, self.data.substation_layout, "gs", "Collection Points"
        )
        plot_cables(ax1, self.data.cable_routes)

        if self.data.lease_poly is not None:
            plot_lease_boundary(ax1, self.data.lease_poly)

        ax1.margins(0.1, 0.1)
        ax1.autoscale_view()

        if self.data.lease_poly is not None:
            annotate_poly(ax1, self.data.lease_poly)

        xlabel = "UTM x [$m$]"
        ylabel = "UTM y [$m$]"

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.ticklabel_format(useOffset=False)
        plt.xticks(rotation=30, ha="right")
        plt.legend(
            handles=[dplot, splot, lplot],
            bbox_to_anchor=(1.05, 1),
            loc=2,
            borderaxespad=0.0,
            numpoints=1,
        )

        plt.title("Electrical Cable Layout")
        plt.tight_layout(rect=(0.0, 0.04, 0.84, 0.96))

        self.fig_handle = plt.gcf()


class ArrayFoundationsPlot(PlotInterface):
    @classmethod
    def get_name(cls):
        """A class method for the common name of the interface.

        Returns:
          str: A unique string
        """

        return "Array Foundations Layout"

    @classmethod
    def declare_inputs(cls):
        """A class method to declare all the variables required as inputs by
        this interface.

        Returns:
          list: List of inputs identifiers

        Example:
          The returned value can be None or a list of identifier strings which
          appear in the data descriptions. For example::

              inputs = ["My:first:variable",
                        "My:second:variable",
                       ]
        """

        input_list = [
            "site.lease_boundary",
            "project.layout",
            "project.substation_layout",
            "project.foundations_component_data",
        ]

        return input_list

    @classmethod
    def declare_optional(cls):
        option_list = ["site.lease_boundary"]

        return option_list

    @classmethod
    def declare_id_map(cls):
        """Declare the mapping for variable identifiers in the data description
        to local names for use in the interface. This helps isolate changes in
        the data description or interface from effecting the other.

        Returns:
          dict: Mapping of local to data description variable identifiers

        Example:
          The returned value must be a dictionary containing all the inputs and
          outputs from the data description and a local alias string. For
          example::

              id_map = {"var1": "My:first:variable",
                        "var2": "My:second:variable",
                        "var3": "My:third:variable"
                       }

        """

        id_map = {
            "lease_poly": "site.lease_boundary",
            "layout": "project.layout",
            "substation_layout": "project.substation_layout",
            "foundations_components": "project.foundations_component_data",
        }

        return id_map

    def connect(self):
        fig = plt.figure()
        ax1 = fig.add_subplot(1, 1, 1, aspect="equal")

        short_layout = {
            key.replace("device", ""): value
            for key, value in self.data.layout.items()
        }

        dplot = plot_point_dict(ax1, short_layout, "k+", "Devices")
        splot = plot_point_dict(
            ax1, self.data.substation_layout, "gs", "Collection Points"
        )

        foundation_marker = {
            "shallowfoundation": "b_",
            "gravity": "r^",
            "pile": "co",
            "suctioncaisson": "mp",
            "directembedment": "y|",
            "drag": "kv",
        }

        foundation_name = {
            "shallowfoundation": "Shallow",
            "gravity": "Gravity",
            "pile": "Pile",
            "suctioncaisson": "Suction Caisson",
            "directembedment": "Direct Embedment",
            "drag": "Drag",
        }

        foundations_locations = self.data.foundations_components[
            ["Type", "UTM X", "UTM Y"]
        ]

        locations_groups = foundations_locations.groupby("Type")
        foundations_handles = []

        for name, group in locations_groups:
            plot_marker = foundation_marker[name]
            plot_name = foundation_name[name]

            coords = group[["UTM X", "UTM Y"]].values
            plot_dict = {str(i): Point(xy) for i, xy in enumerate(coords)}
            fplot = plot_point_dict(
                ax1, plot_dict, plot_marker, plot_name, markersize=8
            )

            foundations_handles.append(fplot)

        if self.data.lease_poly is not None:
            plot_lease_boundary(ax1, self.data.lease_poly)

        ax1.margins(0.1, 0.1)
        ax1.autoscale_view()

        if self.data.lease_poly is not None:
            annotate_poly(ax1, self.data.lease_poly)

        xlabel = "UTM x [$m$]"
        ylabel = "UTM y [$m$]"

        all_handles = [dplot, splot] + foundations_handles

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.ticklabel_format(useOffset=False)
        plt.xticks(rotation=30, ha="right")
        plt.legend(
            handles=all_handles,
            bbox_to_anchor=(1.05, 1),
            loc=2,
            borderaxespad=0.0,
            numpoints=1,
        )

        plt.title("Foundations Layout")
        plt.tight_layout()

        self.fig_handle = plt.gcf()


def plot_point_dict(
    ax,
    layout: dict[str, Point],
    marker,
    label=None,
    annotate=False,
    markersize=None,
    text_weight=None,
    text_size=None,
):
    if text_weight is None:
        text_weight = "normal"
    if text_size is None:
        text_size = "medium"

    x = []
    y = []

    for coords in layout.values():
        x.append(coords.x)
        y.append(coords.y)

    kwargs = {"mew": 2, "markersize": 10}

    if label is not None:
        kwargs["label"] = label
    if markersize is not None:
        kwargs["markersize"] = markersize

    pplot = ax.plot(x, y, marker, **kwargs)

    if not annotate:
        return pplot[0]

    for key, point in layout.items():
        coords = list(point.coords[0])
        ax.annotate(
            key,
            xy=coords[:2],
            xytext=(0, 10),
            xycoords="data",
            textcoords="offset pixels",
            horizontalalignment="center",
            weight=text_weight,
            size=text_size,
        )

    return pplot[0]


def plot_lease_boundary(ax, lease_boundary, padding=None):
    outer_coords = list(lease_boundary.exterior.coords)

    if padding is None:
        lease_boundary = Polygon(outer_coords)
        patch = patch_from_polygon(lease_boundary, ec=GREY, fill=False)

    else:
        inner_boundary = lease_boundary.buffer(-padding)
        inner_coords = list(inner_boundary.exterior.coords)

        # Check if the orientation of the polygons are the same
        if clockwise(*zip(*inner_coords)) == clockwise(*zip(*outer_coords)):
            inner_coords = inner_coords[::-1]

        lease_boundary = Polygon(outer_coords, [inner_coords])
        patch = patch_from_polygon(
            lease_boundary,
            fc=RED,
            fill=True,
            alpha=0.3,
            ls=None,
        )

    ax.add_patch(patch)


def annotate_poly(ax, lease_poly):
    maxy = lease_poly.bounds[3] + 50.0
    centroid = np.array(lease_poly.centroid.coords[0])

    ymin, ymax = ax.get_ylim()

    if maxy > ymax:
        maxy = ymin + 0.95 * (ymax - ymin)

    ax.annotate(
        "Lease Area",
        xy=(centroid[0], maxy),
        horizontalalignment="center",
        verticalalignment="bottom",
        weight="bold",
        size="large",
    )


def plot_cables(ax, cable_routes):
    cables = cable_routes.groupby("Marker")

    xmax = -np.inf
    ymax = -np.inf
    xmin = np.inf
    ymin = np.inf

    for name, cable in cables:
        x = cable["UTM X"]
        y = cable["UTM Y"]

        xmax = max(list(x) + [xmax])
        ymax = max(list(y) + [ymax])
        xmin = min(list(x) + [xmin])
        ymin = min(list(y) + [ymin])

        line = Line2D(x, y)
        ax.add_line(line)


def clockwise(x, y):
    """Use the shoelace formula to determine whether the polygon points are
    defined in a clockwise direction"""
    # https://stackoverflow.com/a/1165943/3215152
    # https://stackoverflow.com/a/19875560/3215152
    if sum(x[i] * (y[i + 1] - y[i - 1]) for i in range(-1, len(x) - 1)) < 0:
        return True
    return False



# -*- coding: utf-8 -*-

#    Copyright (C) 2021-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. moduleauthor:: Mathew Topper <damm_horse@yahoo.co.uk>
"""

import abc
import logging
import os
import pickle
import queue
import sys
import threading
import time
import traceback
from collections import OrderedDict
from copy import deepcopy
from logging import handlers
from math import ceil
from subprocess import Popen
from typing import Any, Sequence

import cma
import numpy as np
from numpy.linalg import norm

from ..files import init_dir

# Convenience import
from .noisehandler import NoiseHandler  # noqa: F401

# Set up logging
module_logger = logging.getLogger(__name__)


class SafeCMAEvolutionStrategy(cma.CMAEvolutionStrategy):
    def plot(self):
        cma.plot(self.logger.name_prefix)


class NormScaler:
    sigma = 1.0
    range_n_sigmas = 3

    def __init__(self, range_min, range_max, x0=None):
        if x0 is None:
            x0 = 0.5 * (range_min + range_max)

        self._scale_factor = _get_scale_factor(
            range_min, range_max, x0, self.sigma, self.range_n_sigmas
        )
        self._x0 = self.scaled(x0)

    @property
    def x0(self):
        return self._x0

    def scaled(self, value):
        return value * self._scale_factor

    def inverse(self, value):
        return value / self._scale_factor


class Counter:
    __metaclass__ = abc.ABCMeta

    def __init__(self, search_dict=None):
        if search_dict is None or not search_dict:
            evaluation = 0
            search_dict = {}
        else:
            evaluation = max(search_dict) + 1

        self._evaluation = evaluation
        self._search_dict = search_dict
        self._lock = threading.Lock()

    @property
    def search_dict(self):
        self._lock.acquire()

        try:
            result = deepcopy(self._search_dict)
        finally:
            self._lock.release()

        return result

    def set_params(self, evaluation, *args):
        params = self._set_params(*args)

        self._lock.acquire()

        try:
            if evaluation in self._search_dict:
                err_str = ("Evaluation {} has already been " "recorded").format(
                    evaluation
                )
                raise ValueError(err_str)

            self._search_dict[evaluation] = params

        finally:
            self._lock.release()

    @abc.abstractmethod
    def _set_params(self, *args):
        """Build a params (probably namedtuple) object to record evaluation."""
        pass

    def get_cost(self, *args):
        self._lock.acquire()
        cost = None

        try:
            for params in self._search_dict.values():
                cost = self._get_cost(params, *args)
                if cost is not None:
                    break

        finally:
            self._lock.release()

        if cost is None:
            cost = False

        return cost

    @abc.abstractmethod
    def _get_cost(self, params, *args):  # pylint: disable=unused-argument
        """Return cost if parameters in params object match input args, else
        return None."""
        pass

    def next_evaluation(self):
        self._lock.acquire()

        try:
            result = self._evaluation
            self._evaluation += 1

        finally:
            self._lock.release()

        return result


class Evaluator:
    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        core,
        base_project,
        root_project_base_name,
        worker_directory,
        restart=False,
        clean_existing_dir=False,
    ):
        self._core = core
        self._base_project = base_project
        self._root_project_base_name = root_project_base_name
        self._worker_directory = worker_directory
        self._counter: Counter = self._init_counter()

        if not restart:
            init_dir(worker_directory, clean_existing_dir)

    @abc.abstractmethod
    def _init_counter(self) -> Counter:
        "Initialise and return the counter stored in the _counter attr"
        pass

    @abc.abstractmethod
    def _get_popen_args(
        self,
        worker_project_path,
        n_evals,
        *args,
    ) -> Sequence[str]:
        "Return the arguments to create a new process thread using Popen"
        pass

    @abc.abstractmethod
    def _get_worker_results(self, evaluation) -> dict[str, Any]:
        """Return the results for the given evaluation as a dictionary that
        must include the key "cost". For constraint violation the cost key
        should be set to np.nan"""
        pass

    @abc.abstractmethod
    def _set_counter_params(
        self, evaluation, worker_project_path, results, flag, n_evals, *args
    ):
        """Update the counter object with new data."""
        pass

    def pre_constraints_hook(self, *args):  # pylint: disable=no-self-use,unused-argument
        """Allows checking of constraints prior to execution. Should return
        True if violated otherwise False"""
        return False

    def _cleanup_hook(self, worker_project_path, flag, results):  # pylint: disable=no-self-use,unused-argument
        """Hook to clean up simulation files as required"""
        pass

    def get_counter_search_dict(self):
        return self._counter.search_dict

    def _iterate(self, results_queue, n_evals, x, *extra):
        previous_cost = self._counter.get_cost(*x)

        if previous_cost:
            results_queue.put((previous_cost,) + extra)
            return

        flag = ""
        evaluation = self._counter.next_evaluation()

        worker_file_root_path = "{}_{}".format(
            self._root_project_base_name, evaluation
        )
        worker_project_name = "{}.dtop".format(worker_file_root_path)
        worker_project_path = os.path.join(
            self._worker_directory, worker_project_name
        )

        results = None
        cost = np.nan

        try:
            self._core.dump_project(self._base_project, worker_project_path)

        except Exception as e:  # pylint: disable=broad-except
            flag = "Fail Send"
            _log_exception(e, flag)

        try:
            popen_args = self._get_popen_args(worker_project_path, n_evals, *x)
            process = Popen(popen_args, close_fds=True)
            exit_code = process.wait()

            if exit_code != 0:
                args_str = ", ".join(popen_args)
                err_str = (
                    "External process failed to open. Arguments were: " "{}"
                ).format(args_str)

                raise RuntimeError(err_str)

        except Exception as e:  # pylint: disable=broad-except
            flag = "Fail Execute"
            _log_exception(e, flag)

        if "Fail" not in flag:
            try:
                results = self._get_worker_results(evaluation)
                cost = results["cost"]

            except Exception as e:  # pylint: disable=broad-except
                flag = "Fail Receive"
                _log_exception(e, flag)

        self._set_counter_params(
            evaluation, worker_project_path, results, flag, n_evals, *x
        )
        self._cleanup_hook(worker_project_path, flag, results)

        results_queue.put((cost,) + extra)

    def __call__(self, q, stop_empty=False):
        """Call the evaluator with a queue.Queue() where index 0 is another
        queue to collect results, index 1 is the number of evaluations for
        noisy cost functions, index 2 is the solution to solve and any extra
        indices are added to the result queue following the calculated cost.
        """

        def check1():
            return not q.empty()

        def check2():
            return True

        if stop_empty:
            check = check1
        else:
            check = check2

        while check():
            item = q.get()
            self._iterate(*item)
            q.task_done()


class Main:
    def __init__(
        self,
        es,
        evaluator,
        scaled_vars,
        x_ops,
        nh=None,
        fixed_index_map=None,
        base_penalty=None,
        num_threads=None,
        maximise=False,
        max_resample_loop_factor=None,
        auto_resample_iterations=None,
    ):
        # Defaults
        if base_penalty is None:
            base_penalty = 1.0
        if num_threads is None:
            num_threads = 1

        self.es = es
        self.nh = nh
        self.evaluator = evaluator
        self._scaled_vars = scaled_vars
        self._x_ops = x_ops
        self._fixed_index_map = fixed_index_map
        self._base_penalty = base_penalty
        self._num_threads = num_threads
        self._maximise = maximise
        self._spare_sols = 1
        self._n_hist = int(10 + 30 * self.es.N / self.es.sp.popsize)
        self._stop = False
        self._sol_penalty = False
        self._dirty_restart = False
        self._sol_feasible = None
        self._max_resample_loops: int
        self._n_record_resample: int
        self._thread_queue: queue.Queue

        self._init_resamples(max_resample_loop_factor, auto_resample_iterations)
        self._init_threads()

    @property
    def stop(self):
        return self._stop

    @property
    def max_resample_loops(self):
        return self._max_resample_loops

    def _init_resamples(
        self, max_resample_loop_factor, auto_resample_iterations
    ):
        if auto_resample_iterations is None:
            if max_resample_loop_factor is None:
                max_resample_loop_factor = 20

            if max_resample_loop_factor < 0:
                err_msg = (
                    "Argument max_resample_loop_factor must be greater "
                    "than or equal to zero"
                )
                raise ValueError(err_msg)

            self._n_record_resample = 0

            n_default = self.es.popsize
            needed_solutions = (1 + self._spare_sols) * n_default
            self._max_resample_loops = int(
                ceil(max_resample_loop_factor * needed_solutions)
            )

            log_msg = ("Setting maximum resamples to " "{}").format(
                self._max_resample_loops
            )
            module_logger.debug(log_msg)

        else:
            if auto_resample_iterations <= 0:
                err_msg = (
                    "Argument auto_resample_iterations must be greater "
                    "than zero"
                )
                raise ValueError(err_msg)

            self._max_resample_loops = 0
            self._n_record_resample = auto_resample_iterations

    def _init_threads(self):
        self._thread_queue = queue.Queue()

        for _ in range(self._num_threads):
            worker = threading.Thread(
                target=self.evaluator, args=(self._thread_queue,)
            )
            worker.daemon = True
            worker.start()

    def next(self):
        if self.es.stop():
            self._stop = True
            return

        if self.nh is None:
            self._next()
        else:
            self._next_nh()

        tolfun = max(self.es.fit.fit) - min(self.es.fit.fit)
        tolfunhist = max(self.es.fit.hist) - min(self.es.fit.hist)

        msg_str = ("Minimum fitness for iteration {}: " "{:.15e}").format(
            self.es.countiter, min(self.es.fit.fit)
        )
        module_logger.info(msg_str)

        msg_str = ("Fitness value range (last iteration): " "{}").format(tolfun)
        module_logger.info(msg_str)

        msg_str = (
            "Minimum fitness value range (last {} iterations): " "{}"
        ).format(self._n_hist, tolfunhist)
        module_logger.info(msg_str)

    def _next(self):
        default, _ = self._get_solutions_costs(self.es)
        self.es.tell(default["solutions"], default["costs"])
        self.es.logger.add()

    def _next_nh(self):
        assert self.nh is not None

        # self._sol_penalty is set by self._get_solutions_costs
        if self.es.countiter == 0 or self._sol_penalty or self._dirty_restart:
            scaled_solutions_extra = None
            last_n_evals = None
            self._dirty_restart = False
        else:
            scaled_solutions_extra = self.nh.ask(self.es.ask)
            last_n_evals = self.nh.last_n_evals

        default, extra = self._get_solutions_costs(
            self.es, scaled_solutions_extra, self.nh.n_evals, last_n_evals
        )

        self.es.tell(default["solutions"], default["costs"])
        self.nh.tell(extra["solutions"], extra["costs"])

        self.es.sigma *= self.nh.sigma_fac
        self.es.countevals += self.nh.evaluations_just_done

        noise = self.nh.get_predicted_noise()

        msg = "Last true noise: {}".format(self.nh.noiseS)
        module_logger.info(msg)

        msg = "Predicted noise: {}".format(noise)
        module_logger.info(msg)

        if abs(noise) <= 1e-12:
            log_noise = 1
        elif noise <= -1e-12:
            log_noise = 0.1
        else:
            log_noise = 10

        self.es.logger.add(more_data=[self.nh.evaluations, log_noise])
        self.nh.prepare(default["solutions"], default["costs"])

    def _get_solutions_costs(
        self,
        asktell,
        scaled_solutions_extra=None,
        n_evals=None,
        n_evals_extra=None,
    ):
        self._sol_penalty = False

        final_solutions = []
        final_costs = []
        final_solutions_extra = []
        final_costs_extra = []

        result_default = {"solutions": final_solutions, "costs": final_costs}
        result_extra = {
            "solutions": final_solutions_extra,
            "costs": final_costs_extra,
        }

        n_default = asktell.popsize
        needed_solutions = (1 + self._spare_sols) * n_default
        xmean = None

        run_solutions = []
        run_descaled_solutions = []
        resample_loops = -1

        while len(run_solutions) < needed_solutions:
            # If the maximum number of resamples is reached apply a
            # penalty value to a new set of solutions
            if (
                self._n_record_resample == 0
                and resample_loops == self._max_resample_loops
            ):
                log_msg = (
                    "Maximum of {} resamples reached. Only {} of {} "
                    "required solutions were found. Applying "
                    "penalty values."
                ).format(
                    self._max_resample_loops,
                    len(run_solutions),
                    needed_solutions,
                )
                module_logger.info(log_msg)

                run_solutions = []
                run_descaled_solutions = []
                final_solutions.extend(asktell.ask(n_default))
                final_costs.extend(self._get_penalty(final_solutions))

                n_default = 0

                # Do not assess noise
                self._sol_penalty = True

                break

            scaled_solutions = asktell.ask(asktell.popsize, xmean)

            # Store xmean for resamples
            if xmean is None:
                xmean = asktell.mean.copy()

            descaled_solutions = self.get_descaled_solutions(scaled_solutions)

            checked_solutions = []
            checked_descaled_solutions = []

            for sol, des_sol in zip(scaled_solutions, descaled_solutions):
                if self.evaluator.pre_constraints_hook(*des_sol):
                    continue

                checked_solutions.append(sol)
                checked_descaled_solutions.append(des_sol)

                solution_str = ", ".join(str(xi) for xi in des_sol)
                log_msg = "Solution found: " + solution_str
                module_logger.debug(log_msg)

            max_sols = needed_solutions - len(run_solutions)
            checked_solutions = checked_solutions[:max_sols]
            checked_descaled_solutions = checked_descaled_solutions[:max_sols]

            run_solutions.extend(checked_solutions)
            run_descaled_solutions.extend(checked_descaled_solutions)

            resample_loops += 1

            if resample_loops > 0 and checked_solutions:
                log_msg = (
                    "{} of {} solutions found after {} resampling " "loops"
                ).format(len(run_solutions), needed_solutions, resample_loops)
                module_logger.debug(log_msg)

        if not self._sol_penalty and resample_loops > 0:
            log_msg = (
                "{} resamples required to generate {} " "solutions"
            ).format(resample_loops, needed_solutions)
            module_logger.info(log_msg)

        if self._n_record_resample > 0:
            if resample_loops > self._max_resample_loops:
                self._max_resample_loops = resample_loops

            self._n_record_resample -= 1

            if self._n_record_resample == 0:
                log_msg = (
                    "Setting maximum resamples to {} following " "iteration {}"
                ).format(self._max_resample_loops, self.es.countiter + 1)
                module_logger.info(log_msg)

        categories = ["default"] * len(run_descaled_solutions)
        all_n_evals = [n_evals] * len(run_descaled_solutions)

        n_extra = 0
        run_solutions_extra = []
        run_descaled_solutions_extra = []
        categories_extra = []
        all_n_evals_extra = []

        if scaled_solutions_extra is not None:
            descaled_solutions_extra = self.get_descaled_solutions(
                scaled_solutions_extra
            )

            for sol, des_sol in zip(
                scaled_solutions_extra, descaled_solutions_extra
            ):
                if self.evaluator.pre_constraints_hook(*des_sol):
                    final_solutions_extra.append(sol)
                    final_costs_extra.append(np.nan)
                else:
                    run_solutions_extra.append(sol)
                    run_descaled_solutions_extra.append(des_sol)
                    categories_extra.append("extra")
                    n_extra += 1

            all_n_evals_extra = [n_evals_extra] * n_extra

        if n_extra + n_default == 0:
            return result_default, result_extra

        run_solutions = run_solutions_extra + run_solutions
        run_descaled_solutions = (
            run_descaled_solutions_extra + run_descaled_solutions
        )
        categories = categories_extra + categories
        all_n_evals = all_n_evals_extra + all_n_evals

        assert (
            len(run_solutions)
            == len(run_descaled_solutions)
            == len(categories)
            == len(all_n_evals)
        )

        # Reuse previous costs if the function is not noisy
        if self.nh is None:
            run_idxs, match_dict = _get_match_process(
                run_descaled_solutions, categories, run_solutions
            )

        else:
            run_idxs = range(len(run_solutions))
            match_dict = {}

        result_queue = queue.Queue()

        for i in range(n_extra + n_default):
            if i not in run_idxs:
                continue

            descaled_sol = run_descaled_solutions[i]
            local_n_evals = all_n_evals[i]
            category = categories[i]
            sol = run_solutions[i]

            item: list[Any] = [result_queue]
            item.append(local_n_evals)
            item.append(descaled_sol)
            item.extend([i, category, sol])

            self._thread_queue.put(item)

        store_results = {}
        min_i = 0
        next_i = n_extra + n_default
        results_found = 0

        while results_found < n_extra + n_default:
            cost, i, category, sol = result_queue.get()

            if i in match_dict:
                all_i = [(i, category, sol)] + match_dict[i]
                all_cost = [cost] * (1 + len(match_dict[i]))
            else:
                all_i = [(i, category, sol)]
                all_cost = [cost]

            for k, kcost in zip(all_i, all_cost):
                i, category, sol = k
                store_results[i] = (category, sol, kcost)

            for check_i in range(min_i, next_i):
                if check_i not in store_results:
                    continue
                if check_i == min_i:
                    min_i += 1

                result = store_results.pop(check_i)

                if result[0] == "extra":
                    extra_cost = result[2]
                    if self._maximise:
                        extra_cost *= -1

                    final_solutions_extra.append(result[1])
                    final_costs_extra.append(extra_cost)
                    results_found += 1
                    continue

                assert result[0] == "default"

                sol = result[1]
                sol_cost = result[2]

                if np.isnan(sol_cost):
                    if next_i == len(run_descaled_solutions):
                        log_msg = (
                            "Maximum number of retry solutions "
                            "reached. Applying penalty value."
                        )
                        module_logger.info(log_msg)

                        sol_cost = self._get_penalty([sol])[0]

                    else:
                        if next_i not in run_idxs:
                            next_i += 1
                            continue

                        run_descaled_sol = run_descaled_solutions[next_i]
                        run_category = categories[next_i]
                        run_sol = run_solutions[next_i]

                        item = [result_queue]
                        item.append(n_evals)
                        item.append(run_descaled_sol)
                        item.extend([next_i, run_category, run_sol])

                        self._thread_queue.put(item)

                        next_i += 1
                        continue

                if self._maximise:
                    sol_cost *= -1
                final_solutions.append(sol)
                final_costs.append(sol_cost)

                results_found += 1

        self._thread_queue.join()

        return result_default, result_extra

    def get_descaled_solutions(self, scaled_solutions):
        descaled_solutions = []

        for solution in scaled_solutions:
            new_solution = [
                scaler.inverse(x)
                for x, scaler in zip(solution, self._scaled_vars)
            ]
            new_solution = [
                op(x) if op is not None else x
                for x, op in zip(new_solution, self._x_ops)
            ]

            descaled_solutions.append(new_solution)

        if self._fixed_index_map is not None:
            ordered_index_map = OrderedDict(
                sorted(self._fixed_index_map.items(), key=lambda t: t[0])
            )

            for solution in descaled_solutions:
                for idx, val in ordered_index_map.items():  # pylint: disable=no-member
                    solution.insert(idx, val)

        return descaled_solutions

    def _get_penalty(self, sols):
        if self._sol_feasible is None:
            if self.es.countiter == 0:
                err_msg = (
                    "There are no feasible solutions to "
                    "use for penalty calculation. Problem "
                    "is likely ill-posed."
                )
                raise RuntimeError(err_msg)

            self._sol_feasible = self.es.best.x.copy()

        costs = [
            self._base_penalty + norm(sol - self._sol_feasible) for sol in sols
        ]

        return costs

    def get_max_resample_factor(self):
        if self._n_record_resample > 0:
            return "auto{}".format(self._n_record_resample)

        return self._max_resample_loops


def init_evolution_strategy(
    x0,
    low_bound,
    high_bound,
    integer_variables=None,
    max_simulations=None,
    popsize=None,
    timeout=None,
    tolfun=None,
    logging_directory=None,
):
    opts = {"bounds": [low_bound, high_bound]}  # ,
    #            'verbose': -3}

    if integer_variables is not None:
        opts["integer_variables"] = integer_variables

    if max_simulations is not None:
        opts["maxfevals"] = max_simulations

    if popsize is not None:
        opts["popsize"] = popsize

    if timeout is not None:
        opts["timeout"] = timeout

    if tolfun is not None:
        opts["tolfun"] = tolfun

    if logging_directory is not None:
        opts["verb_filenameprefix"] = logging_directory + os.sep

    es = SafeCMAEvolutionStrategy(x0, NormScaler.sigma, opts)

    return es


def dump_outputs(worker_directory, es, evaluator, nh=None):
    counter_dict = evaluator.get_counter_search_dict()

    es_path = os.path.join(worker_directory, "saved-cma-object.pkl")
    counter_dict_path = os.path.join(
        worker_directory, "saved-counter-search-dict.pkl"
    )

    with open(es_path, "wb") as f:
        pickle.dump(es, f, -1)

    with open(counter_dict_path, "wb") as f:
        pickle.dump(counter_dict, f, -1)

    if nh is None:
        return

    nh_path = os.path.join(worker_directory, "saved-nh-object.pkl")
    with open(nh_path, "wb") as f:
        pickle.dump(nh, f, -1)


def load_outputs(worker_directory):
    es_path = os.path.join(worker_directory, "saved-cma-object.pkl")
    nh_path = os.path.join(worker_directory, "saved-nh-object.pkl")
    counter_dict_path = os.path.join(
        worker_directory, "saved-counter-search-dict.pkl"
    )

    es = pickle.load(open(es_path, "rb"))
    counter_dict = pickle.load(open(counter_dict_path, "rb"))

    if os.path.isfile(nh_path):
        nh = pickle.load(open(nh_path, "rb"))
    else:
        nh = None

    return es, counter_dict, nh


def set_TimedRotatingFileHandler_rollover(timeout=None):
    # If theres no timeout choose a big number
    if timeout is None:
        timeout = sys.maxsize

    logger = logging.Logger.manager.loggerDict["dtocean_core"]
    if not isinstance(logger, logging.Logger):
        return

    for handler in logger.handlers:
        if isinstance(handler, handlers.TimedRotatingFileHandler):
            handler.interval = timeout
            handler.rolloverAt = handler.computeRollover(int(time.time()))


def _get_scale_factor(range_min, range_max, x0, sigma, n_sigmas):
    if x0 < range_min or x0 > range_max:
        err_str = "x0 must lie between range_min and range_max"
        raise ValueError(err_str)

    if sigma <= 0:
        err_str = "sigma must be positive"
        raise ValueError(err_str)

    if n_sigmas % 1 != 0 or int(n_sigmas) <= 0:
        err_str = "n_sigmas must be a positive whole number"
        raise ValueError(err_str)

    max_half_range = max(range_max - x0, x0 - range_min)
    half_scaled_range = sigma * int(n_sigmas)

    return half_scaled_range / max_half_range


def _get_match_process(values, *args):
    def expand_args(x, args):
        if not args:
            return x

        result = [x]

        for arg in args:
            result.append(arg[x])

        return tuple(result)

    match_dict = {}
    all_matches = []

    for i in range(len(values) - 1):
        if i in all_matches:
            continue

        key = i
        base = values[i]
        match_list = []

        for j in range(i + 1, len(values)):
            test = values[j]

            if all([x == y for x, y in zip(base, test)]):
                match_list.append(j)

        all_matches.extend(match_list)

        if match_list:
            match_list = [expand_args(k, args) for k in match_list]
            match_dict[key] = match_list

    process_set = set(range(len(values))) - set(all_matches)

    return list(process_set), match_dict


def _log_exception(e, flag):
    module_logger.debug(flag)
    module_logger.debug(e)

    exc_type, exc_value, exc_traceback = sys.exc_info()
    msg_strs = traceback.format_exception(exc_type, exc_value, exc_traceback)
    msg_str = "".join(msg_strs)
    module_logger.debug(msg_str)



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from copy import deepcopy

from mdo_engine.boundary.interface import MapInterface


class Tool(MapInterface):
    """The base abstract class for all tool classes"""

    def __init__(self):
        super(Tool, self).__init__()
        self._config = {}

    @abc.abstractmethod
    def configure(self, *args, **kwargs):
        """The configure method is collect information required for executing
        the tool.
        """
        pass

    def get_config(self):
        return deepcopy(self._config)

    def set_config(self, config_dict):
        self._config = config_dict
